/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    filterPatchDictionary.cpp
** AUTHOR(S):   Olga Russakovsky <olga@cs.stanford.edu>
**                based on trainObjectDetector.cpp by Steve Gould
** DESCRIPTION:
**   Trims the dictionary down to a more manageable size; eliminates
**   features that are not very informative on the training set
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <sys/types.h>
#include <vector>
#include "svlDevel.h"

#if defined(_WIN32)||defined(WIN32)||defined(__WIN32__)
#include "win32/dirent.h"
#else
#include <dirent.h>
#endif

#include "cv.h"
#include "cxcore.h"
#include "highgui.h"

#include "svlBase.h"
#include "svlML.h"
#include "svlVision.h"

using namespace std;

// Prototypes -----------------------------------------------------------------

vector<vector<double> > readFromCache(const char* dirName, int maxImages, bool binary);
void updateMinMIincrease(float *minMIincrease, const int max_index, const vector<bool> &selected,
			 const svlClassifierResponseCache *responses, const float *MI);

// Thread functions

struct createResponsesArgs {
  createResponsesArgs(svlClassifierResponseCache *c, float *cMI,
		      const int idx,
		      const vector<vector<double> > *pos,
		      const vector<vector<double> > *neg) :
    cur_cache(c), cur_MI(cMI), index(idx),
    posSamples(pos), negSamples(neg) {}
  
  svlClassifierResponseCache *cur_cache;
  float *cur_MI;
  const int index;
  const vector<vector<double> > *posSamples;
  const vector<vector<double> > *negSamples;
  
};

void *createResponses(void *voidArgs, unsigned tid) {
  createResponsesArgs *args = (createResponsesArgs *) voidArgs;
  
  const int i = args->index;
  const vector<vector<double> > &posSamples = *(args->posSamples);
  const vector<vector<double> > &negSamples = *(args->negSamples);

  svlClassifierResponseList responses;
  responses.reserve(posSamples.size() + negSamples.size());
  for (unsigned j = 0; j < posSamples.size(); j++)
    insertScore(responses, posSamples[j][i], true);
  for (unsigned j = 0; j < negSamples.size(); j++)
    insertScore(responses, negSamples[j][i], false);

  *(args->cur_cache) = thresholdResponsesForMaxMI(responses, args->cur_MI);

  //SVL_LOG(SVL_LOG_VERBOSE, "Thresholded patch " << i);

  return NULL;
}


struct updateMIincreaseArgs {
  updateMIincreaseArgs(float *r, const float *m,
			const svlClassifierResponseCache *c1,
			const svlClassifierResponseCache *c2) :
    result(r), orig_MI(m), responses1(c1), responses2(c2) {}
  
  float *result;
  const float *orig_MI;
  const svlClassifierResponseCache *responses1;
  const svlClassifierResponseCache *responses2;
};
  
void *updateMIincrease(void *voidArgs, unsigned tid) {

  updateMIincreaseArgs *args = (updateMIincreaseArgs *) voidArgs;

  float *result = args->result;
  const float orig_MI = *(args->orig_MI);
  const svlClassifierResponseCache &responses1 = *(args->responses1);
  const svlClassifierResponseCache &responses2 = *(args->responses2);

  *result = min(*result, computeJointMI(responses1, responses2) - orig_MI);  

  return NULL;
}


// Main -----------------------------------------------------------------------

void usage()
{
    cerr << SVL_USAGE_HEADER << endl;
    cerr << "USAGE: ./filterPatchDictionary <dictionary> <+ cache dir> <- cache dir>" << endl;
    cerr << "OPTIONS:" << endl
	 << "  -o <filename>       :: output filename for the new dictionary" << endl
	 << "  -cc <threshold>     :: threshold for cross-correlation of patches (default: 0.9)" << endl
	 << "  -F <threshold>      :: threshold for F-score of patches (default: 0.75)" << endl
	 << "  -MI <threshold>     :: instead of the above, use the mutual information of the " << endl
	 << "                         features to make a decision; threshold is the lowest acceptable " << endl
	 << "                         increase in MI for a patch to be added (default: 0)" << endl
	 << "  -minSize <n>        :: minimum number of features expected (overrides -F and -MI, not -cc)" << endl
	 << "  -maxSize <n>        :: maximum number of features expected" << endl
	 << "  -maxInstances <n>   :: maximum training images for each class (default: 1000)" << endl
	 << "  -binary             :: the cache dirs are really names of binary files" << endl
         << "  -v                  :: verbose" << endl
	 << endl;
}

bool patch_scores_sort_decreasing(pair<float, int> i, pair<float, int> j) {
  return i.first > j.first;
}

int main(int argc, char *argv[])
{
  // read commandline parameters
  const int NUM_REQUIRED_PARAMETERS = 3;
  
  const char *posCacheDir = NULL;
  const char *negCacheDir = NULL;
  
  const char *dictionaryFilename = NULL;
  const char *outputFilename = NULL;
  int maxTrainingInstances = 1000;
  double CCthreshold = 0.9;
  double Fthreshold = 0.75;
  int minNumPatches = 0;
  int maxNumPatches = INT_MAX;
  double MIthreshold = 0;
  
  bool binary = false;
  bool useMI = false;
  
  char **args = argv + 1;
  while (--argc > NUM_REQUIRED_PARAMETERS) {
    if (!strcmp(*args, "-maxInstances")) {
      maxTrainingInstances = atoi(*(++args)); argc--;
    } else if (!strcmp(*args, "-o")) {
      outputFilename = *(++args);          argc--;
    } else if (!strcmp(*args, "-cc")) {
      CCthreshold = atof(*(++args));       argc--;
    } else if (!strcmp(*args, "-F")) {
      Fthreshold = atof(*(++args));         argc--;
    } else if (!strcmp(*args, "-MI")) {
      MIthreshold = atof(*(++args));         argc--;
      useMI = true;
    } else if (!strcmp(*args, "-minSize")) {
      minNumPatches = atoi(*(++args));      argc--;
    } else if (!strcmp(*args, "-maxSize")) {
      maxNumPatches = atoi(*(++args));      argc--;
    } else if (!strcmp(*args, "-binary")) {
      binary = true;
    } else if (!strcmp(*args, "-v")) {
      svlLogger::setLogLevel(SVL_LOG_VERBOSE);
    } else {
      SVL_LOG(SVL_LOG_FATAL, "unrecognized option " << *args);
    }
    args++;
  }
  
  if (argc != NUM_REQUIRED_PARAMETERS) {
    usage();
    return -1;
  }

  dictionaryFilename = args[0];
  posCacheDir = args[1];
  negCacheDir = args[2];
  
  // read the dictionary
  svlPatchDictionary dict;
  dict.read(dictionaryFilename);
  
  // load positive and negative training samples from cache
  SVL_LOG(SVL_LOG_VERBOSE, "Reading positive samples...");
  vector<vector<double> > posSamples = svlReadExamplesFromCache(posCacheDir, maxTrainingInstances, binary);
  SVL_LOG(SVL_LOG_VERBOSE, "..." << posSamples.size() << " samples read");
  SVL_LOG(SVL_LOG_VERBOSE, "Reading negative samples...");
  vector<vector<double> > negSamples = svlReadExamplesFromCache(negCacheDir, maxTrainingInstances, binary);
  SVL_LOG(SVL_LOG_VERBOSE, "..." << negSamples.size() << " samples read");
  
  unsigned numPatches = dict.numFeatures();
  
  // make sure all samples were correctly cached
  for (int j = (int)posSamples.size() - 1; j >= 0; j--) {
    if (posSamples[j].size() != numPatches) {
      posSamples.erase(posSamples.begin() + j);
      SVL_LOG(SVL_LOG_WARNING, "Had to erase posSample " << j << ", incorrect num features");
    }
  }
  
  for (int j = (int)negSamples.size() - 1; j >= 0; j--) {
    if (negSamples[j].size() != numPatches) {
      negSamples.erase(negSamples.begin() + j);
      SVL_LOG(SVL_LOG_WARNING, "Had to erase negSample " << j << ", incorrect num features");
    }
  }
  
  // these are indexed by the original ordering, corresponding to the 
  // dictionary entries
  vector<bool> selected(numPatches, false);
  int numSelected = 0;
  
  // this is the original implementation, which uses the F-score of each patch and a rough
  // approximation of the correlation between two patches
  if (!useMI) {
    
    vector<pair<float, int> > patchFscores(numPatches);
    
    // analyze each of the dictionary features in turn
    // to see how good it is at discriminating between +/- examples
    // (note: the +/- examples are weighted s.t. each set contributed 1/2 the weight
    // even though the number of each might be different)
    for (unsigned i = 0; i < numPatches; i++) {
      
      svlClassifierResponseList c;
      for (unsigned j = 0; j < posSamples.size(); j++)
	insertScore(c, posSamples[j][i], true);
      for (unsigned j = 0; j < negSamples.size(); j++)
	insertScore(c, negSamples[j][i], false);
      
      normalizeScores(c);
      svlPRcurve p;
      createPRcurve(c, p);
      patchFscores[i] = make_pair(getHighestFscore(p), i);
    }
    
    // sort the patches in order of Fscore, highest first
    stable_sort(patchFscores.begin(), patchFscores.end(), patch_scores_sort_decreasing);
    
    // these are indexed by the current ordering of patches
    vector<bool> availableForSelection(numPatches, true);
    
    // iterate over patches in order of decreasing Fscores
    for (unsigned i = 0; (i < numPatches) && (numSelected < maxNumPatches) &&
	   (patchFscores[i].first > Fthreshold || numSelected < minNumPatches); i++) {
      if (availableForSelection[i]) {
	SVL_LOG(SVL_LOG_VERBOSE, "Analyzing patch " << i << " of " << numPatches
		<< ", with value " << patchFscores[i].first);
	int index1 = patchFscores[i].second;
	selected[index1] = true;
	numSelected++;
	
	int num = 0;
	for (unsigned j = i+1; j < numPatches; j++) {
	  if (availableForSelection[j]) {
	    int index2 = patchFscores[j].second;
	    if (dict.entriesTooCorrelated(index1, index2, CCthreshold)) {
	      availableForSelection[j] = false;
	      num++;
	    }
	  } 
	}
	SVL_LOG(SVL_LOG_VERBOSE, "Deleted " << num << " patches as a result");
      }
    }
  } // endif !useMI
  
  else { // useMI
    
    // this is straight out of:
    // Michel Vidal-Naquet and Shimon Ullman. Object Recognition with Informative Features
    // and Linear Classification, ICCV 2003
    // (although it's nothing more than standard feature selection)

    //SVL_LOG(SVL_LOG_VERBOSE, "Starting to threshold " << numPatches << " patches");
    
    svlClassifierResponseCache *responses = new svlClassifierResponseCache[numPatches];
    float *MI = new float[numPatches];      
    
    svlThreadPool threadPool(1);
    threadPool.start();
    
    // used for storing the arguments and later cleaning them up
    void **threadArgs = new void*[numPatches];
    for (unsigned i = 0; i < numPatches; i++) {
      threadArgs[i] = new createResponsesArgs(&(responses[i]), &(MI[i]),
					      i, &posSamples, &negSamples);
      threadPool.addJob(createResponses, threadArgs[i]);
    }
    
    threadPool.finish();

    for (unsigned i = 0; i < numPatches; i++)
      delete (createResponsesArgs *)(threadArgs[i]);

    float max_MI = -1.0f;
    int max_index = -1;
    for (unsigned i = 0; i < numPatches; i++) {
      if (MI[i] > max_MI) {
	max_MI = MI[i];
	max_index = i;
      }
    }
    
    // select the patch with the highest mutual information
    selected[max_index] = true;
    numSelected++;
    SVL_LOG(SVL_LOG_VERBOSE, "Selected patch " << max_index << ", with MI of " << MI[max_index]);
    
    // stores the minimum jointMI between each patch and any of the patches that have
    // been selected so far
    float *minMIincreaseWithSelected = new float[numPatches];
    for (unsigned i = 0; i < numPatches; i++) {
      minMIincreaseWithSelected[i] = FLT_MAX;
    }

    while (numSelected < maxNumPatches) {
      // update the minimums based on the last patch chosen
      threadPool.start();
      
      for (unsigned i = 0; i < numPatches; i++) {
	threadArgs[i] = NULL;
	if (selected[i]) continue;

	threadArgs[i] = new updateMIincreaseArgs(&(minMIincreaseWithSelected[i]),
						 &(MI[max_index]), &(responses[max_index]),
						 &(responses[i]));
	threadPool.addJob(updateMIincrease, threadArgs[i]);
      }
      
      threadPool.finish();
      
      for (unsigned i = 0; i < selected.size(); i++)
	if (threadArgs[i])
	  delete (updateMIincreaseArgs *)(threadArgs[i]);
      
      // find the next best patch to add
      float MIincrease = -1.0f;
      max_index = -1;
      for (unsigned i = 0; i < numPatches; i++) {
	if (selected[i]) continue;
	
	float increase = minMIincreaseWithSelected[i];
	if (increase > MIincrease) {
	  MIincrease = increase;
	  max_index = i;
	}
      }
      
      if (MIincrease < MIthreshold && numSelected >= minNumPatches)
	break;
      
      selected[max_index] = true;
      numSelected++;
      SVL_LOG(SVL_LOG_VERBOSE, "Selected patch " << max_index << ", with MI increase of " << MIincrease);
    }
    
    for (unsigned i = 0; i < numPatches; i++)
      deleteClassifierResponseCache(responses[i]);
    delete[] responses;
    delete[] minMIincreaseWithSelected;
    delete[] MI;
    delete[] threadArgs;
  }
  
  dict.filterEntries(selected);
  
  SVL_LOG(SVL_LOG_MESSAGE, "Trimmed dictionary to " << dict.numFeatures() << " patches");
  
  dict.write(outputFilename);
  
  return 0;
}







