/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    segImageExtractEdgeFeatures.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
** DESCRIPTION:
**  Code for extracting edge features from image segment boundaries.
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>

#include "svlBase.h"
#include "svlVision.h"

using namespace std;

#define WINDOW_NAME "segImageExtractEdgeFeatures"

void usage()
{
    cerr << SVL_USAGE_HEADER << "\n";
    cerr << "USAGE: ./segImageExtractEdgeFeatures [OPTIONS] <imageDir>\n";
    cerr << "OPTIONS:\n"
         << "  -o <directory>         :: output directory (default: current)\n"
         << "  -imgExt <ext>          :: image file extension (default: .jpg)\n"
         << "  -segExt <ext>          :: over-segmentation file extension (default: .seg)\n"
         << "  -segDir <dir>          :: over-segmentation file directory (default: <imageDir>)\n"
         << "  -featuresExt <ext>     :: output feature file extension (default: .edgeFeatures.txt)\n"
         << "  -edgePairsExt <ext>    :: edge pairs file extension (default: .pairs.txt)\n"
         << "  -includeColorDiff      :: include colour difference (default: false)\n"
         << "  -includeBias           :: include bias feature (default: false)\n"
         << "  -skipEdgePairs         :: skip extraction of edge pairs\n"
         << "  -v                     :: verbose\n"
         << "  -x                     :: visualize feature extraction\n";
    cerr << "<imageDir> must contain images <base>.jpg and <regionsDir> must contain\n"
         << "corresponding over-segmentations <base>.seg.\n"
         << endl;
}

int main(int argc, char *argv[])
{
    svlCodeProfiler::enabled = true;
    svlCodeProfiler::tic(svlCodeProfiler::getHandle("main"));

    // read commandline parameters
    const int NUM_REQUIRED_PARAMETERS = 1;

    const char *imgDir = NULL;
    const char *segDir = NULL;
    const char *outputDir = ".";
    const char *imgExt = ".jpg";
    const char *segExt = ".seg";
    const char *featuresExt = ".edgeFeatures.txt";
    const char *edgePairsExt = ".pairs.txt";
    bool bIncludeColorDiff = false;
    bool bIncludeBias = false;
    bool bSkipEdgePairs = false;
    bool bVisualize = false;

    char **args = argv + 1;
    while (--argc > NUM_REQUIRED_PARAMETERS) {
        if (!strcmp(*args, "-o")) {
            outputDir = *(++args); argc -= 1;
        } else if (!strcmp(*args, "-imgExt")) {
            imgExt = *(++args); argc -= 1;
        } else if (!strcmp(*args, "-segExt")) {
            segExt = *(++args); argc -= 1;
        } else if (!strcmp(*args, "-segDir")) {
            segDir = *(++args); argc -= 1;
        } else if (!strcmp(*args, "-featuresExt")) {
            featuresExt = *(++args); argc -= 1;
        } else if (!strcmp(*args, "-edgePairsExt")) {
            edgePairsExt = *(++args); argc -= 1;
        } else if (!strcmp(*args, "-includeColorDiff")) {
            bIncludeColorDiff = true;
        } else if (!strcmp(*args, "-includeBias")) {
            bIncludeBias = true;
        } else if (!strcmp(*args, "-skipEdgePairs")) {
            bSkipEdgePairs = true;
        } else if (!strcmp(*args, "-v")) {
            svlLogger::setLogLevel(SVL_LOG_VERBOSE);
        } else if (!strcmp(*args, "-x")) {
            bVisualize = true;
        } else {
            SVL_LOG(SVL_LOG_FATAL, "unrecognized option " << *args);
        }
        args++;
    }

    if (argc != NUM_REQUIRED_PARAMETERS) {
        usage();
        return -1;
    }

    imgDir = args[0];
    if (segDir == NULL) segDir = imgDir;

    if (bVisualize) {
        cvNamedWindow(WINDOW_NAME, 1);
    }

    // load images
    SVL_LOG(SVL_LOG_MESSAGE, "Reading images from " << imgDir << "...");
    vector<string> baseNames = svlDirectoryListing(imgDir, imgExt, false, false);
    SVL_LOG(SVL_LOG_MESSAGE, "...read " << baseNames.size() << " images");

    // extract features
    int hImageLoad = svlCodeProfiler::getHandle("imageLoad");
    int hComputeFeatures = svlCodeProfiler::getHandle("computeEdgeFeatures");
    int hFeatureSave = svlCodeProfiler::getHandle("featureSave");
    int hPairsSave = svlCodeProfiler::getHandle("edgePairsSave");
    
    SVL_LOG(SVL_LOG_MESSAGE, "Processing images...");
    for (int i = 0; i < (int)baseNames.size(); i++) {
        // instantiate svlSegImage
        SVL_LOG(SVL_LOG_VERBOSE, "Loading image " << baseNames[i]);
        string imgName = string(imgDir) + string("/") + baseNames[i] + string(imgExt);
        string segName = string(segDir) + string("/") + baseNames[i] + string(segExt);

        svlCodeProfiler::tic(hImageLoad);
        svlSegImageBase img(imgName.c_str(), segName.c_str());
        svlCodeProfiler::toc(hImageLoad);
        SVL_LOG(SVL_LOG_VERBOSE, "..." << toString(*img.image()));

        // show segmentation
        if (bVisualize) {
            IplImage *debugImage = img.visualize();
            img.colorBoundaries(debugImage, 0xff, 0x00, 0x00);
            cvShowImage(WINDOW_NAME, debugImage);
            cvWaitKey(50);
            cvReleaseImage(&debugImage);
        }
        
        // compute features
        SVL_LOG(SVL_LOG_VERBOSE, "...computing features");
        svlCodeProfiler::tic(hComputeFeatures);
        svlEdgeFeatures featureCalculator(img.image());
        vector<vector<double> > x = featureCalculator.computeFeatures(img.getEdgels());
        if (bIncludeColorDiff) {
            svlRegionFeatures regionFeatureCalculator;
            regionFeatureCalculator.setOption("noTexture", true);
            regionFeatureCalculator.setOption("noGeometry", true);
            regionFeatureCalculator.setOption("noLocation", true);
            regionFeatureCalculator.setOption("includeVariance", false);
            regionFeatureCalculator.setOption("includeSkewness", false);
            regionFeatureCalculator.setOption("includeKurtosis", false);
            
            vector<vector<double> > xSeg = regionFeatureCalculator.computeFeatures(img);
            vector<pair<int, int> > adjList = img.getAdjacencyList();
            for (int j = 0; j < (int)adjList.size(); j++) {
                int segA = adjList[j].first;
                int segB = adjList[j].second;                
                x[j].push_back(fabs(xSeg[segA][0] - xSeg[segB][0]));
                x[j].push_back(fabs(xSeg[segA][1] - xSeg[segB][1]));
                x[j].push_back(fabs(xSeg[segA][2] - xSeg[segB][2]));
            }                        
        }
        svlCodeProfiler::toc(hComputeFeatures);
        SVL_LOG(SVL_LOG_VERBOSE, "..." << x.size() << " feature vectors computed");
        
        // save features
        svlCodeProfiler::tic(hFeatureSave);
        string outputName = string(outputDir) + string("/") + baseNames[i] + string(featuresExt);
        ofstream ofs(outputName.c_str());
        assert(!ofs.fail());
        for (unsigned j = 0; j < x.size(); j++) {
            if (bIncludeBias) ofs << "1 ";
            for (unsigned k = 0; k < x[j].size(); k++) {
                if (k != 0) ofs << " ";
                ofs << x[j][k];
            }
            ofs << "\n";
        }
        ofs.close();
        svlCodeProfiler::toc(hFeatureSave);

        // save edge pairs
        if (!bSkipEdgePairs) {
            svlCodeProfiler::tic(hPairsSave);
            vector<pair<int, int> > adjList = img.getAdjacencyList();

            string outputName = string(outputDir) + string("/") + baseNames[i] + string(edgePairsExt);
            ofstream ofs(outputName.c_str());
            assert(!ofs.fail());
            for (int i = 0; i < (int)adjList.size(); i++) {
                ofs << adjList[i].first << " " << adjList[i].second << "\n";
            }
            ofs.close();
            svlCodeProfiler::toc(hPairsSave);
        }        
    }
    SVL_LOG(SVL_LOG_MESSAGE, "...done");
    
    // free memory
    cvDestroyAllWindows();
    svlCodeProfiler::toc(svlCodeProfiler::getHandle("main"));
    svlCodeProfiler::print(cerr);
    return 0;
}
