/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlCudaConvolution.cpp
** AUTHOR(S):   Paul Baumstarck <pbaumstarck@stanford.edu>
** DESCRIPTION:
**  Handles NVIDIA Cuda memory interface.
**
*****************************************************************************/

#include "svlCudaConvolution.h"


svlCudaConvolution::svlCudaConvolution() {}

svlCudaConvolution::~svlCudaConvolution() {}

// Pad in width and height.
pair<int,int> fullPad(const IplImage *patch, float* &pp, float* &pmask)
{
	// Bin patch into a proper size and then convert to linear memory.
	int patch_w = -1, patch_h = -1;
	if ( patch->width <= 4 && patch->height <= 4 ) {
		patch_w = 4;
		patch_h = 4;
	} else if ( patch->width <= 8 && patch->height <= 8 ) {
		patch_w = 8;
		patch_h = 8;
	} else if ( patch->width <= 8 && patch->height <= 16 ) {
		patch_w = 8;
		patch_h = 16;
	} else if ( patch->width <= 16 && patch->height <= 8 ) {
		patch_w = 16;
		patch_h = 8;
	} else if ( patch->width <= 16 && patch->height <= 16 ) {
		patch_w = 16;
		patch_h = 16;
	} else {
		assert(false);
	}

	// Construct patch padded to above size.
	pp = (float*)malloc(sizeof(float)*patch_w*patch_h);
	pmask = (float*)malloc(sizeof(float)*patch_w*patch_h);
	float *ptr = pp, *ptrm = pmask;
	int r, c;
	// Do on-patch rows
	for ( r=0; r<patch->height; ++r ) {
		// Do on-patch columns
		for ( c=0; c<patch->width; ++c ) {
			*ptr++ = (float)CV_IMAGE_ELEM(patch, unsigned char, r, c);
			*ptrm++ = 1.0f;
		}
		// Zeralize remainder columns
		for ( ; c<patch_w; ++c ) {
			*ptr++ = 0.0f;
			*ptrm++ = 0.0f;
		}
	}
	// Zeralize remainder rows
	for ( ; r<patch_h; ++r ) {
		for ( c=0; c<patch_w; ++c ) {
			*ptr++ = 0.0f;
			*ptrm++ = 0.0f;
		}
	}

#if 0
	if ( patch->width == 5 && patch->height == 4 ) {
		// Dump patch for debugging.
		FILE *fo = fopen("patch.out","wb");
		fwrite(pp,sizeof(float),patch_w*patch_h,fo);
		fclose(fo);
		fo = fopen("mask.out","wb");
		fwrite(pmask,sizeof(float),patch_w*patch_h,fo);
		fclose(fo);
		cout << "5 by 4; out\n"; exit(-1);
	}
#endif
	return pair<int,int>(patch_w,patch_h);
}

// Pad in width.
pair<int,int> halfPad(const IplImage *patch, float* &pp, float* &pmask)
{
	// Bin patch into a proper size and then convert to linear memory.
	int patch_w = -1, patch_h = patch->height;
	if ( patch->width <= 4 ) {
		patch_w = 4;
	} else if ( patch->width <= 8 ) {
		patch_w = 8;
	} else if ( patch->width <= 16 ) {
		patch_w = 16;
	} else {
		assert(false);
	}

	// Construct patch padded to above size.
	pp = (float*)malloc(sizeof(float)*patch_w*patch_h);
	pmask = (float*)malloc(sizeof(float)*patch_w*patch_h);
	float *ptr = pp, *ptrm = pmask;
	int r, c;
	// Do on-patch rows
	for ( r=0; r<patch->height; ++r ) {
		// Do on-patch columns
		for ( c=0; c<patch->width; ++c ) {
			*ptr++ = (float)CV_IMAGE_ELEM(patch, unsigned char, r, c);
			*ptrm++ = 1.0f;
		}
		// Zeralize remainder columns
		for ( ; c<patch_w; ++c ) {
			*ptr++ = 0.0f;
			*ptrm++ = 0.0f;
		}
	}
	// Zeralize remainder rows
	for ( ; r<patch_h; ++r ) {
		for ( c=0; c<patch_w; ++c ) {
			*ptr++ = 0.0f;
			*ptrm++ = 0.0f;
		}
	}

#if 0
	// Dump patch for debugging.
	FILE *fo = fopen("patch.out","wb");
	fwrite(pp,sizeof(float),patch_w*patch_h,fo);
	fclose(fo);
	fo = fopen("mask.out","wb");
	fwrite(pmask,sizeof(float),patch_w*patch_h,fo);
	fclose(fo);
#endif
	return pair<int,int>(patch_w,patch_h);
}

// Convolve a pitch with a patch described as an IplImage and output to another pitch.
void svlCudaConvolution::normCCorr(const svlCudaPitch *pitch_in, const IplImage *patch, svlCudaPitch *pitch_out,
								   int img_w, int img_h, float patch_mean, float patch_std, svlCudaCoveringGrid *cudagrid)
{
	float *pp, *pmask;
	pair<int,int> wh = fullPad(patch, pp, pmask);
	int patch_w = wh.first, patch_h = wh.second;
	// Run convolution.
	svlCudaNormCCorr2d(*pitch_in, *pitch_out, pp, pmask, patch_w, patch_h, patch->height*patch->width, patch_mean, patch_std, img_w, img_h, cudagrid);
	free(pp);
	free(pmask);
}

// Convolve a pitch with a patch described as an IplImage and output to another pitch -- v2.
void svlCudaConvolution::normCCorr_v2(const svlCudaPitch *pitch_in, const IplImage *patch, svlCudaPitch *pitch_out,
									  int img_w, int img_h, float patch_mean, float patch_std)
{
	float *pp, *pmask;
	pair<int,int> wh = halfPad(patch, pp, pmask);
	int patch_w = wh.first, patch_h = wh.second;

	// Run convolution.
	svlCudaNormCCorr2d_v2(*pitch_in, *pitch_out, pp, pmask, patch_w, patch_h, patch->height*patch->width, patch_mean, patch_std, img_w, img_h);
	free(pp);
	free(pmask);
}

// The above, but use v3.
void svlCudaConvolution::normCCorr_v3(const svlCudaPitch *pin, const IplImage *patch, svlCudaPitch *pout, svlCudaPitch *integ, svlCudaPitch *integ2,
									  int img_w, int img_h, float patch_mean, float patch_std)
{
	float *pp, *pmask;
	pair<int,int> wh = halfPad(patch, pp, pmask);
	int patch_w = wh.first, patch_h = wh.second;

	// Run convolution.
	svlCudaNormCCorr2d_v3(pin, pout, pp, pmask, integ, integ2, patch_w, patch_h, patch->height*patch->width, patch_mean, patch_std, img_w, img_h);
	free(pp);
	free(pmask);
}

// Convolve a pitch with a patch described as an IplImage and output to another pitch.
void svlCudaConvolution::normCCorr2(const svlCudaPitch *pitch_in, const IplImage *patch,
									svlCudaPitch *pitch_out, svlCudaPitch *temp1, svlCudaPitch *temp2,
									int img_w, int img_h, float patch_mean, float patch_std)
{
	float *pp, *pmask;
	pair<int,int> wh = fullPad(patch, pp, pmask);
	int patch_w = wh.first, patch_h = wh.second;

	// Run convolution.
	svlCudaNormCCorr2d_twostep(*pitch_in, *pitch_out, *temp1, *temp2,
		pp, pmask, patch_w, patch_h, patch->height*patch->width, patch_mean, patch_std, img_w, img_h);
	free(pp);
	free(pmask);
}

pair<int,int> svlCudaConvolution::padPatch(const IplImage *patch, float* &pad_patch, float* &pad_mask)
{
	return fullPad(patch, pad_patch, pad_mask);
}

// Returns the size in linear memory "patch" will take when padded for device convolution.
int svlCudaConvolution::linearPatchSize(IplImage *patch)
{
	pair<int,int> ret = paddedPatchSize(patch);
	return ret.first * ret.second;
}

// Get padded patch size.
pair<int,int> svlCudaConvolution::paddedPatchSize(IplImage *patch)
{
	pair<int,int> ret;
	ret.second = patch->height;
	if ( patch->width <= 4 ) {
		ret.first = 4;
	} else if ( patch->width <= 8 ) {
		ret.first = 8;
	} else if ( patch->width <= 16 ) {
		ret.first = 16;
	} else {
		assert(false);
	}
	return ret;
}

// Pad the patch to device convolution size and store it and its mask in the provided memory.
void svlCudaConvolution::padPatchWithMask(IplImage *patch, float *mem)
{
	// Bin patch into a proper size and then convert to linear memory.
	pair<int,int> pad_wh = paddedPatchSize(patch);
	int sz = pad_wh.first * pad_wh.second;

	// Construct patch padded to above size.
	float *pdata = (float*)malloc(sizeof(float)*sz*2);
	float *ptr = pdata, *ptrm = pdata + sz;
	int r, c;
	// Do on-patch rows
	for ( r=0; r<patch->height; ++r ) {
		// Do on-patch columns
		for ( c=0; c<patch->width; ++c ) {
			*ptr++ = (float)CV_IMAGE_ELEM(patch, unsigned char, r, c);
			*ptrm++ = 1.0f;
		}
		// Zeralize remainder columns
		for ( ; c<pad_wh.first; ++c ) {
			*ptr++ = 0.0f;
			*ptrm++ = 0.0f;
		}
	}
	// Zeralize remainder rows
	for ( ; r<pad_wh.second; ++r ) {
		for ( c=0; c<pad_wh.first; ++c ) {
			*ptr++ = 0.0f;
			*ptrm++ = 0.0f;
		}
	}
	memcpy(mem, pdata, sizeof(float)*sz*2);
	free(pdata);
}


