/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlCudaConvolution.h
** AUTHOR(S):   Paul Baumstarck <pbaumstarck@stanford.edu>
** DESCRIPTION:
**  Handles NVIDIA Cuda memory interface.
**
*****************************************************************************/

#pragma once

#ifndef __SVL_CUDA_CONVOLUTION_H
#define __SVL_CUDA_CONVOLUTION_H

#include <iostream>
#include <vector>

#include "cv.h"
#include "cxcore.h"

#include "svlCudaCommon.h"
#include "svlCudaMemHandler.h"
using namespace std;

class svlCudaConvolution {
public:
	svlCudaConvolution();
	virtual ~svlCudaConvolution();

	// 2D convolution
	template<class V> static void conv2d(const svlCudaPitch *pitch_in, const IplImage *patch, svlCudaPitch *pitch_out, int img_w = -1, int img_h = -1);
	template<class V> static void conv2d_v2(const svlCudaPitch *pitch_in, const IplImage *patch, svlCudaPitch *pitch_out, int img_w = -1, int img_h = -1);

	// Normalized cross correlation function
	static void normCCorr(const svlCudaPitch *pitch_in, const IplImage *patch, svlCudaPitch *pitch_out,
		int img_w = -1, int img_h = -1, float patch_mean = 0.0f, float patch_std = 1.0f, svlCudaCoveringGrid *cudagrid = NULL);
	static void normCCorr2(const svlCudaPitch *pitch_in, const IplImage *patch, svlCudaPitch *pitch_out,
		svlCudaPitch *temp1, svlCudaPitch *temp2, int img_w, int img_h, float patch_mean, float patch_std);
	static void normCCorr_v2(const svlCudaPitch *pitch_in, const IplImage *patch, svlCudaPitch *pitch_out, int img_w, int img_h, float patch_mean, float patch_std);
	static void normCCorr_v3(const svlCudaPitch *pin, const IplImage *patch, svlCudaPitch *pout, svlCudaPitch *integ, svlCudaPitch *integ2,
		int img_w, int img_h, float patch_mean, float patch_std);
	// Using pre-device-cached patches.
	inline static void normCCorr_v2Cached(const svlCudaPitch *pin, svlCudaPitch *pout, float *ppatch, float *pmask,
		int patch_w, int patch_h, int patch_numel, float patch_mean, float patch_std, int img_w = -1, int img_h = -1) {
		svlCudaNormCCorr2d_v2(*pin, *pout, ppatch, pmask, patch_w, patch_h, patch_numel, patch_mean, patch_std, img_w, img_h, true);
	}
	inline static void normCCorr_v3Cached(const svlCudaPitch *pin, svlCudaPitch *pout, float *ppatch, float *pmask, svlCudaPitch *integ, svlCudaPitch *integ2,
		int patch_w, int patch_h, int patch_numel, float patch_mean, float patch_std, int img_w = -1, int img_h = -1) {
		svlCudaNormCCorr2d_v3(pin, pout, ppatch, pmask, integ, integ2, patch_w, patch_h, patch_numel, patch_mean, patch_std, img_w, img_h, true);
	}
	// Using pre-padded patches and masks.
	inline static void normCCorrPadded(const svlCudaPitch *pitch_in, svlCudaPitch *pitch_out, float *patch, float *mask,
		int patch_w, int patch_h, int patch_real_numel, float patch_mean = 0.0f, float patch_std = 1.0f, int img_w = -1, int img_h = -1,
		svlCudaCoveringGrid *cudagrid = NULL) {
		svlCudaNormCCorr2d(*pitch_in, *pitch_out, patch, mask, patch_w, patch_h, patch_real_numel, patch_mean, patch_std, img_w, img_h, cudagrid);
	}
	// Get pre-computed window sums and sum-squareds.
	inline static void windowSumSumSq(const svlCudaPitch *pin, svlCudaPitch *sum, svlCudaPitch *sum2, int win_w, int win_h, int img_w = -1, int img_h = -1) {
		svlCudaWindowSumSumSq(pin, sum, sum2, win_w, win_h, img_w, img_h);
	}
	// Normalize an image using aligned sum and sum-squared images.
	inline static void patchNormalize(svlCudaPitch *pout, svlCudaPitch *sum, svlCudaPitch *sum2, float patch_mean, float patch_std, int patch_N, int img_w = -1, int img_h = -1) {
		svlCudaPatchNormalize(pout, sum, sum2, patch_mean, patch_std, patch_N, img_w = -1, img_h = -1);
	}

	// Patch-sized window normalization
	inline static void windowNormalize(svlCudaPitch *pitch_in, const svlCudaPitch *integral, const svlCudaPitch *integral2,
		float patch_mean, float patch_std, int patch_w, int patch_h, int img_w = -1, int img_h = -1) {
		svlCudaWindowNormalize(*pitch_in, *integral, *integral2, patch_mean, patch_std, patch_w, patch_h, img_w, img_h);
	}
	// Normalize with off-by-one integral images.
	inline static void windowNormalize1(svlCudaPitch *pitch_in, const svlCudaPitch *integral, const svlCudaPitch *integral2,
		float patch_mean, float patch_std, int patch_w, int patch_h, int img_w = -1, int img_h = -1) {
		svlCudaWindowNormalize1(*pitch_in, *integral, *integral2, patch_mean, patch_std, patch_w, patch_h, img_w, img_h);
	}
	inline static void windowNormalize1DimSub(svlCudaPitch *pitch_in, const svlCudaPitch *integral, const svlCudaPitch *integral2,
		float patch_mean, float patch_std, int patch_w, int patch_h, int img_w = -1, int img_h = -1, int dim = 128) {
		svlCudaWindowNormalize1DimSub(*pitch_in, *integral, *integral2, patch_mean, patch_std, patch_w, patch_h, img_w, img_h, dim);
	}
	inline static void windowNormalizeRow(svlCudaPitch *pin, svlCudaPitch *pout, int patch_w, int img_w = -1, int img_h = -1, int dim = 128) {
		svlCudaWindowNormalizeRow(pin, pout, patch_w, img_w, img_h, dim);
	}

	// Other filtering functions.
	inline static void sobel(svlCudaPitch *pin, svlCudaPitch *edge, svlCudaPitch *temp1, svlCudaPitch *temp2, int img_w = -1, int img_h = -1) {
		svlCudaSobel(pin, edge, temp1, temp2, img_w, img_h);
	}

	// Miscellaneous functions
	static pair<int,int> padPatch(const IplImage *patch, float* &pad_patch, float* &pad_mask);
	static pair<int,int> paddedPatchSize(IplImage *patch);
	static int linearPatchSize(IplImage *patch);
	static void padPatchWithMask(IplImage *patch, float *mem);
};

// Convolve a pitch with a patch described as an IplImage and output to another pitch.
template<class V>
void svlCudaConvolution::conv2d(const svlCudaPitch *pitch_in, const IplImage *patch, svlCudaPitch *pitch_out, int img_w, int img_h)
{
	// Bin patch into a proper size and then convert to linear memory.
	int patch_w = -1, patch_h = -1;
	//printf("In (%d,%d) ...\n", patch->width, patch->height);
	if ( patch->width <= 4 && patch->height <= 4 ) {
		patch_w = 4;
		patch_h = 4;
	} else if ( patch->width <= 8 && patch->height <= 8 ) {
		patch_w = 8;
		patch_h = 8;
	} else if ( patch->width <= 8 && patch->height <= 16 ) {
		patch_w = 8;
		patch_h = 16;
	} else if ( patch->width <= 16 && patch->height <= 8 ) {
		patch_w = 16;
		patch_h = 8;
	} else if ( patch->width <= 16 && patch->height <= 16 ) {
		patch_w = 16;
		patch_h = 16;
	} else {
		assert(false);
	}
	//printf("Out: (%d,%d)\n", patch_w, patch_h);

	// Construct padded patch.
	float *pp = (float*)malloc(sizeof(float)*patch_w*patch_h);
	float *ptr = pp;
	int r, c;
	// Do on-patch rows
	for ( r=0; r<patch->height; ++r ) {
		// Do on-patch columns
		for ( c=0; c<patch->width; ++c )
			*ptr++ = (float)CV_IMAGE_ELEM(patch, V, r, c);
		// Zeralize remainder columns
		for ( ; c<patch_w; ++c )
			*ptr++ = 0.0f;
	}
	// Zeralize remainder rows
	for ( ; r<patch_h; ++r ) {
		for ( c=0; c<patch_w; ++c )
			*ptr++ = 0.0f;
	}

#if 0
	// Dump patch for debugging.
	FILE *fo = fopen("patch_debug.out","wb");
	fwrite(pp,sizeof(float),patch_w*patch_h,fo);
	fclose(fo);
#endif
	// Run convolution.
	svlCudaConvolve2d(*pitch_in, *pitch_out, pp, patch_w, patch_h, img_w, img_h);
	free(pp);
}

// Convolve a pitch with a patch described as an IplImage and output to another pitch.
template<class V>
void svlCudaConvolution::conv2d_v2(const svlCudaPitch *pitch_in, const IplImage *patch, svlCudaPitch *pitch_out, int img_w, int img_h)
{
	// Bin patch into a proper size and then convert to linear memory.
	int patch_w = -1, patch_h = patch->height;
	if ( patch->width <= 4 ) {
		patch_w = 4;
	} else if ( patch->width <= 8 ) {
		patch_w = 8;
	} else if ( patch->width <= 16 ) {
		patch_w = 16;
	} else {
		assert(false);
	}

	// Construct padded patch.
	float *pp = (float*)malloc(sizeof(float)*patch_w*patch_h);
	float *ptr = pp;
	int r, c;
	// Do on-patch rows
	for ( r=0; r<patch->height; ++r ) {
		// Do on-patch columns
		for ( c=0; c<patch->width; ++c )
			*ptr++ = (float)CV_IMAGE_ELEM(patch, V, r, c);
		// Zeralize remainder columns
		for ( ; c<patch_w; ++c )
			*ptr++ = 0.0f;
	}

#if 0
	// Dump patch for debugging.
	FILE *fo = fopen("patch_debug.out","wb");
	fwrite(pp,sizeof(float),patch_w*patch_h,fo);
	fclose(fo);
#endif
	// Run convolution.
	svlCudaConvolve2d_v2(*pitch_in, *pitch_out, pp, patch_w, patch_h, img_w, img_h);
	free(pp);
}

#endif

