/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlCudaSlidingWindowDetector.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**              Sid Batra <sidbatra@stanford.edu>
**              Olga Russakovsky <olga@cs.stanford.edu>
**              Ian Goodfellow <ia3n@cs.stanford.edu>
**
*****************************************************************************/

#include <cassert>
#include <iostream>
#include <fstream>
#include <vector>
#include <sys/types.h>
#include <limits>

#include "cxcore.h"
#include "ml.h"

#include "svlBase.h"
#include "svlVision.h"
#include "svlCudaSlidingWindowDetector.h"

#if defined(_WIN32)||defined(WIN32)||defined(__WIN32__)
#define _CRT_SECURE_NO_DEPRECATE
#include "win32/dirent.h"
#else
#include <dirent.h>
#endif

#define VERBOSE
#undef DEBUG
//#define DEBUG

// Cuda running options.
#define RAND_ACCESS			// Use linear indexing to compact results on the GPU before sending back.
#define DEVICE_INTEGRALS	// Define this to use on-device integral images
#define ROW_INTEGS
#define HIGHLANDER_DEBUG 0

using namespace std;

svlCudaMemHandler *cm;

static inline bool isSampleValid(vector<double> &sample) {
	for (unsigned i = 0; i < sample.size(); i++)
		if (isinf(sample[i]) || isnan(sample[i])) return false;
	return true;
}

// Dumps an IplImage to a named file.
template <class C>
void iplimage2file(const char *file, const IplImage *im)
{
	FILE *fo = fopen(file,"wb");
	fwrite(&im->width,sizeof(int),1,fo);
	fwrite(&im->height,sizeof(int),1,fo);
	C gf;
	for ( int i=0; i<im->height; ++i ) {
		for ( int j=0; j<im->width; ++j ) {
			//gf = float(im->imageData[ i*im->widthStep + j ]);
			gf = CV_IMAGE_ELEM(im, C, i, j);
			fwrite(&gf,sizeof(C),1,fo);
		}
	}
	fclose(fo);
}


// Constructors/Destructors --------------------------------------------------

svlCudaSlidingWindowDetector::svlCudaSlidingWindowDetector(int max_img_w, int max_img_h, int max_channels, const char *name, int w, int h)
	: svlSlidingWindowDetector(name, w, h), max_channels(max_channels)
{
	// Allocate memory.
//#ifndef OLD_SCHOOL
	_IMGS.reserve(max_channels);
	/*_CONVS.reserve(max_channels);
	_TEMPS1.reserve(max_channels);
	_TEMPS2.reserve(max_channels);*/
	_INTEGS.reserve(max_channels);
	_INTEGS2.reserve(max_channels);
	//int big_w = svlCudaMath::iAlignUp(max_img_w,128),
	//	big_h = svlCudaMath::iAlignUp(max_img_h,128);
	//if ( big_w == max_img_w )
	//	++big_w; // Have to add one for integral images.
	//if ( big_h == max_img_h )
	//	++big_h; // Have to add one for integral images.
	_CONV = cm->allocPitch(max_img_w, max_img_h);
	_TEMP1 = cm->allocPitch(max_img_w + 1, max_img_h);
	_TEMP2 = cm->allocPitch(max_img_w + 1, max_img_h); // Plus one as they could possibly used for device integral images.
	for ( int i=0; i<max_channels; ++i ) {
		/*_CONVS.push_back( cm->allocPitch(max_img_w, max_img_h) );
		_TEMPS1.push_back( cm->allocPitch(max_img_w, max_img_h) );
		_TEMPS2.push_back( cm->allocPitch(max_img_w, max_img_h) );*/
		_IMGS.push_back( cm->allocPitch(max_img_w, max_img_h) );
		_INTEGS.push_back( cm->allocPitch(max_img_w + 1, max_img_h + 1) );
		_INTEGS2.push_back( cm->allocPitch(max_img_w + 1, max_img_h + 1) );
//#ifndef DEVICE_INTEGRALS
//		_IMGS.push_back( cm->allocPitch(max_img_w, max_img_h) );
//		_INTEGS.push_back( cm->allocPitch(max_img_w+1, max_img_h+1) );
//		_INTEGS2.push_back( cm->allocPitch(max_img_w+1, max_img_h+1) );
//#else
//		/*_IMGS.push_back( cm->allocPitch(big_w, big_h) );
//		_INTEGS.push_back( cm->allocPitch(big_w, big_h) );
//		_INTEGS2.push_back( cm->allocPitch(big_w, big_h) );*/

		// Sanity check: allocated values.
		/*printf("Got: %d %d %d %d %d %d\n", _IMGS.back(), _CONVS.back(), _TEMPS1.back(), _TEMPS2.back(), _INTEGS.back(), _INTEGS2.back());
		printf(" %x\n", cm->pitchData(_IMGS.back()));
		printf(" %x\n", cm->pitchData(_CONVS.back()));
		printf(" %x\n", cm->pitchData(_TEMPS1.back()));
		printf(" %x\n", cm->pitchData(_TEMPS2.back()));
		printf(" %x\n", cm->pitchData(_INTEGS.back()));
		printf(" %x\n", cm->pitchData(_INTEGS2.back()));*/
//#endif
	}
	//printf("Got: %d %d %d %d %d %d\n", _IMGS[0], _CONVS[0], _TEMPS1[0], _TEMPS2[0], _INTEGS[0], _INTEGS2[0]);
//#else
//
//	_CONV = cm->allocPitch(max_img_w, max_img_h);
//	_TEMP1 = cm->allocPitch(max_img_w, max_img_h);
//	_TEMP2 = cm->allocPitch(max_img_w, max_img_h);
//#ifndef DEVICE_INTEGRALS
//	_IMG = cm->allocPitch(max_img_w, max_img_h);
//	_INTEG = cm->allocPitch(max_img_w+1, max_img_h+1);
//	_INTEG2 = cm->allocPitch(max_img_w+1, max_img_h+1);
//#else
//	int big_w = svlCudaMath::iAlignUp(max_img_w,256),
//		big_h = svlCudaMath::iAlignUp(max_img_h,256);
//	_IMG = cm->allocPitch(big_w, big_h);
//	if ( big_w == max_img_w )
//		++big_w; // Have to add one for integral images.
//	if ( big_h == max_img_h )
//		++big_h; // Have to add one for integral images.
//	_INTEG = cm->allocPitch(big_w, big_h);
//	_INTEG2 = cm->allocPitch(big_w, big_h);
//#endif
//
//#endif
	//printf("Got: %d %d %d %d %d %d\n", _IMG, _CONV, _TEMP1, _TEMP2, _INTEG, _INTEG2);
	
	// Make sure we have a valid NULL lodged in pointer variables.
	cachedLine = NULL;
	dtree = NULL;
}

svlCudaSlidingWindowDetector::svlCudaSlidingWindowDetector(int max_img_w, int max_img_h, int max_channels, const svlCudaSlidingWindowDetector& c)
	: svlSlidingWindowDetector(c), max_channels(max_channels)
{
	assert(false);
	// 'Cause I just have to copy over the stuff above, but I don't feel like it right now.
}

svlCudaSlidingWindowDetector::~svlCudaSlidingWindowDetector()
{
	if (_featureVector != NULL) {
		cvReleaseMat(&_featureVector);
	}

	// Free all device memory.
	for ( vector<svlCudaPitch*>::iterator it = _IMGS.begin(); it != _IMGS.end(); ++it )
		cm->freePitch(*it);
	/*for ( vector<svlCudaPitch*>::iterator it = _CONVS.begin(); it != _CONVS.end(); ++it )
		cm->freePitch(*it);
	for ( vector<svlCudaPitch*>::iterator it = _TEMPS1.begin(); it != _TEMPS1.end(); ++it )
		cm->freePitch(*it);
	for ( vector<svlCudaPitch*>::iterator it = _TEMPS2.begin(); it != _TEMPS2.end(); ++it )
		cm->freePitch(*it);*/
	for ( vector<svlCudaPitch*>::iterator it = _INTEGS.begin(); it != _INTEGS.end(); ++it )
		cm->freePitch(*it);
	for ( vector<svlCudaPitch*>::iterator it = _INTEGS2.begin(); it != _INTEGS2.end(); ++it )
		cm->freePitch(*it);

	cm->freePitch(_CONV);
	cm->freePitch(_TEMP1);
	cm->freePitch(_TEMP2);

	if (_weakResponses != NULL) {
		cvReleaseMat(&_weakResponses);
	}
}

// Compute a binary mask (in floats) 
extern void computeCudaCoveringGrid(vector<CvPoint>& locations, int block_x, int block_y, int dx, int dy, svlCudaCoveringGrid &cgrid);

#ifdef OLD_SCHOOL
// Computes interesting window locations for image GPU'ed at _IMG.
void svlCudaSlidingWindowDetector::createInterestingWindowLocations(IplImage *image, vector<CvPoint>& locations) const
{
	int maxX = image->width - _windowWidth;
	int maxY = image->height - _windowHeight;

	if ( maxX <= 0 || maxY <= 0 )
		return;

	// Copy over image to GPU.
	if ( !cm->pitchHasCached((void*)image, _IMG) ) {
		cm->IplToGPU<unsigned char>(image, _IMG);
	}
#if 0
	iplimage2file<unsigned char>("image.out", image);

	// Image (ix-0) to min (ix-1) and max (ix-2).
	svlCudaMath::minMax32x32( (*cudamem)[0], (*cudamem)[1], (*cudamem)[2], image->width, image->height);
	
	pitch2file("min4.out", cudamem, 1);
	pitch2file("max4.out", cudamem, 2);

	// Perform <max -= min;> (<(ix-2) -= (ix-1);>).
	svlCudaMath::subtract( (*cudamem)[2], (*cudamem)[1], image->width, image->height);
	// Decimate by factors to retrieve only needed data: ix-2 to ix-3.
	svlCudaMath::decimate( (*cudamem)[2], (*cudamem)[3], _delta_x, _delta_y, image->width, image->height);

	pitch2file("decimate.out", cudamem, 3);
#else
	// Image IMG to min CONV and max TEMP1.
	//svlCudaMath::minMax32x32( (*cudamem)[0], (*cudamem)[1], (*cudamem)[2], image->width, image->height);
	svlCudaMath::minMax32x32(_IMG, _CONV, _TEMP1, image->width, image->height);

	// Perform <max -= min;> (TEMP1 -= CONV).
	//svlCudaMath::subtract( (*cudamem)[2], (*cudamem)[1], image->width, image->height);
	svlCudaMath::subtract(_TEMP1, _CONV, image->width, image->height);
	// Decimate by factors to retrieve only needed data: TEMP1 to TEMP2.
	//svlCudaMath::decimate( (*cudamem)[2], (*cudamem)[3], _delta_x, _delta_y, image->width, image->height);
	svlCudaMath::decimate(_TEMP1, _TEMP2, _delta_x, _delta_y, image->width, image->height);
#endif

	// Copy out result.
	const int sub_width = image->width / _delta_x;
	const int sub_height = image->height / _delta_y;
	float *back = (float*)malloc(sizeof(float)*sub_width*sub_height);
	cm->floatFromGPU(back, sub_width, sub_height, _TEMP2);

	// Get min_delta of image uniformity conditions.
	float min_delta = 32.0f; // Default of IPL_DEPTH_8U
	if ( image->depth == IPL_DEPTH_32F ) {
		min_delta = 0.125f;
	}
	locations.reserve((maxX + _delta_x) * (maxY + _delta_y) / (_delta_x * _delta_y));
	int ix, iy = 0;
	for (int y=0; y <= maxY; y += _delta_y) {
		ix = 0;
		for (int x=0; x <= maxX; x += _delta_x) {
			if ( back[ iy*sub_width + ix ] > min_delta ) {
				locations.push_back(cvPoint(x, y));
			}
			++ix;
		}
		++iy;
	}
	locations.resize(locations.size()); // Throw away extra pre-allocated things.
	free(back);
}
#else
// Computes interesting window locations for image saved at pitch_ix.
void svlCudaSlidingWindowDetector::createInterestingWindowLocations(svlCudaPitch *img, int img_w, int img_h, vector<CvPoint>& locations) const
{
	int maxX = img_w - _windowWidth;
	int maxY = img_h - _windowHeight;
	if ( maxX <= 0 || maxY <= 0 )
		return;

	// Image IMG to min CONV and max TEMP1.
	svlCudaPitch *minp = _CONV, *maxp = _TEMP1; // Alias owned variables.
	svlCudaMath::minMax32x32(img, minp, maxp, img_w, img_h);

	// Perform <max -= min;>.
	svlCudaMath::subtract(maxp, minp, img_w, img_h);
	// Decimate by factors to retrieve only needed data; maxp to TEMP2.
	svlCudaMath::decimate(maxp, _TEMP2, _delta_x, _delta_y, img_w, img_h);

	// Copy out result.
	const int sub_width = img_w / _delta_x;
	const int sub_height = img_h / _delta_y;
	float *back = (float*)malloc(sizeof(float)*sub_width*sub_height);
	cm->floatFromGPU(back, sub_width, sub_height, _TEMP2);

	// Get min_delta of image uniformity conditions.
	// TODO paulb: Fix making this dependent on image depth by piping in an image or something.
	float min_delta = 32.0f; // Default of IPL_DEPTH_8U
	/*if ( image->depth == IPL_DEPTH_32F ) {
		min_delta = 0.125f;
	}*/
	locations.reserve((maxX + _delta_x) * (maxY + _delta_y) / (_delta_x * _delta_y));
	int ix, iy = 0;
	for (int y=0; y <= maxY; y += _delta_y) {
		ix = 0;
		for (int x=0; x <= maxX; x += _delta_x) {
			if ( back[ iy*sub_width + ix ] > min_delta ) {
				locations.push_back(cvPoint(x, y));
			}
			++ix;
		}
		++iy;
	}
	locations.resize(locations.size()); // Throw away extra pre-allocated things.
	free(back);
}
#endif

#ifndef OLD_SCHOOL
// Classifies a multi-channel image.
svlObject2dFrame svlCudaSlidingWindowDetector::cudaClassifyImage(const vector<IplImage *>& images, int outputOffset)
{
	// Set globals and allocate temporaries.
	//use_integral_images = false; // Set use of integral images.
	svlCudaCoveringGrid *cudagrid = NULL; // Set no covering grid available.

	// Assert that we have not been given any NULL images or images of different size.
	assert(images[0] != NULL);
	for ( unsigned i=1; i<images.size(); ++i ) {
		assert(images[i] != NULL);
		assert(images[i]->width == images[0]->width);
		assert(images[i]->height == images[0]->height);
	}

	// Code profiling handles.
	const int prof00 = svlCodeProfiler::getHandle("cSWD::classify");
	const int extract_handler = svlCodeProfiler::getHandle("cSWD::extract");
	const int read_features = svlCodeProfiler::getHandle("cSWD::read features");
	const int inner_cuda = svlCodeProfiler::getHandle("cSWD::inner cuda");
	const int feat_upkeep = svlCodeProfiler::getHandle("cSWD::feat upkeep");
	const int win_locs = svlCodeProfiler::getHandle("cSWD::window locs");
	const int eval_model = svlCodeProfiler::getHandle("cSWD::eval model");
	const int scale_handle = svlCodeProfiler::getHandle("cSWD::scale image");

	svlCodeProfiler::tic(prof00);

	// Find out how many images we will need in this batch.
	map<int,int> channels2pitches; // This will map from channels selected to device pitch indices.
	// Loop through dictionary to see which channels will be used.
	int gi = 0;
	for ( vector<svlIntensityPatch*>::const_iterator p = dict._intEntries.begin(); p != dict._intEntries.end(); ++p ) {
		if ( channels2pitches.count( (*p)->channel ) == 0 ) {
			channels2pitches[(*p)->channel] = gi++;
		}
	}
	// Copy all images over to GPU.
	for ( int i=0; i<int(images.size()); ++i ) {
		if ( images[i]->depth == IPL_DEPTH_8U ) {
			cm->IplToGPU<unsigned char>( images[i], _IMGS[ channels2pitches[i] ] );
		} else {
			printf("Mysterious depth?: %d\n", images[i]->depth); assert(false);
		}
	}
#ifdef ROW_INTEGS
	int dim = 256; // Integration dimension (must be power of 2).
#endif

#if HIGHLANDER_DEBUG
	static int xena_outer, xena_inner; char c[32];
#endif

	// Objects to return.
	svlObject2dFrame objects;
	float scoreThreshold = probabilityToScore((float)_threshold);
	int numChannels = (int)images.size();
	double scale = 1.0;

	SVL_LOG(SVL_LOG_MESSAGE, "Running classifier on " << numChannels << "-channel image");

	// Create int32 image for storing indices.
	IplImage* ixImage = cvCreateImage(cvSize(images[0]->width, images[0]->height), IPL_DEPTH_32S, 1);
	// Features that they are not allocated afresh every time (soon to be optimized away since all feature calculation will be on the device.
	Tsamples features;
	unsigned neededSize = outputOffset + unsigned(dict.numFeatures());
	if ( outputOffset > 0 ) {
		cout << "Possibly have to transfer features to device before outputOffset > 0\n"; assert(false);
	}

	// Variables for using device-side decision trees.
	bool use_cuda_dtree = dtree != NULL;
	int feature_stride = -1, aligned_n_trees;
	bool native_pow_2 = false; // True if rounded-up stride is equal to original stride, which dictates using cumsum-style summation instead of scan-style.
	svlCudaLine *dfeatures, *ixline, *locline;
	if ( use_cuda_dtree ) {
		// Find feature stride for device-side decision tree computation (max of feature size and tree size, rounded up to nearest factor of 2).
		aligned_n_trees = svlCudaMath::iAlignUpPow2( dtree->getNTrees() );
		feature_stride = max( aligned_n_trees, dict.numFeatures() );
		native_pow_2 = dtree->getNTrees() == feature_stride;
		printf("Aligned: %d; feature stride: %d; pow2: %d\n", aligned_n_trees, feature_stride, native_pow_2?1:0);
	}

	//vector<IplImage*> scaleds;

	// Set current working image scale.
	int scaled_w = images[0]->width, scaled_h = images[0]->height;
	// Iterate over all scales
	while ( scaled_w >= _windowWidth && scaled_h >= _windowHeight )
	{
#if HIGHLANDER_DEBUG
		// Output images in use.
		for ( int i=0; i<int(images.size()); ++i ) {
			sprintf(c,"img_%d_%d.out",xena_outer,i);
			cm->pitch2file(c, _IMGS[i]);
		}
#endif

#ifdef ROW_INTEGS
		// Row-integrate images.
		for ( int i=0; i<(int)images.size(); ++i ) {
			svlCudaMath::rowIntegralBlock(_IMGS[i], _INTEGS[i], _INTEGS2[i], scaled_w, scaled_h, dim);
#if HIGHLANDER_DEBUG
			sprintf(c,"integ_%d_%d.out",xena_outer,i);
			cm->pitch2file(c, _INTEGS[i]);
			sprintf(c,"integ2_%d_%d.out",xena_outer,i);
			cm->pitch2file(c, _INTEGS2[i]);
#endif
		}
#endif

		// Get interesting locations.
		svlCodeProfiler::tic(win_locs);
		vector<CvPoint> locations;
		createInterestingWindowLocations(_IMGS[0], scaled_w, scaled_h, locations); // Get locations from channel 0.
		svlCodeProfiler::toc(win_locs);

#if HIGHLANDER_DEBUG
		// Output all locations.
		sprintf(c,"locs_%d.out",xena_outer);
		FILE *floc = fopen(c,"wb");
		for ( int i=0; i<(int)locations.size(); ++i ) {
			fwrite(&locations[i].x,sizeof(int),1,floc);
			fwrite(&locations[i].y,sizeof(int),1,floc);
		}
		fclose(floc);
#endif

		if ( !locations.empty() ) {

			// Compute features.
			svlCodeProfiler::tic(extract_handler);
			svlCodeProfiler::tic(feat_upkeep);

			if ( !use_cuda_dtree ) {
				// Allocate memory for return vectors.
				if ( features.size() != locations.size() )
					features.resize(locations.size());
				// Resize features's inner vectors to needed size.
				for ( vector<vector<double> >::iterator it = features.begin(); it != features.end(); ++it ) {
					if ( it->size() < neededSize)
						it->resize(neededSize);
				}
			} else {
				// Allocate feature space on device.
				//printf("Got %d locs\n", int(locations.size()));
				int gi = feature_stride * int(locations.size());
				dfeatures = cm->allocLine(gi);
				//printf("Allocated %d on device\n", gi);
				ixline = cm->allocLine( 2*int(locations.size()) );
				locline = cm->allocLine( 2*int(locations.size()) );
				svlCudaMemHandler::CvPoint2IntLine(locations, locline, true); // Cache locations in ixline.
			}

			svlCodeProfiler::toc(feat_upkeep);

			//svlDepthPatchDefinition::_locationsize = _locationsize;

			// TODO paulb: Put this all back in later.
			//Check if any channels are depth or abs
			/*vector<bool> isDepthChannel(images.size(),false);
			vector<bool> isAbsChannel(images.size(), false);*/
			/*for (unsigned i = 0; i < _entries.size(); i++)
			{
				if (_entries[i]->type == SVL_DEPTH_PATCH)
					isDepthChannel[ _entries[i]->channel ] = true;

				if (_entries[i]->type == SVL_ABS_PATCH)
					isAbsChannel[ _entries[i]->channel ] = true;
			}
			for (unsigned i = 0; i < images.size(); i++)
			{
				if (isDepthChannel[i])
					normalizeDepthPatches(images[i], locations);

				if (isAbsChannel[i])
					computeAbsResponses(images[i]);
			}*/

			// Compute feature responses from dictionary.
			// TODO paulb: Convert these to linear memory on the GPU.
			float *results = (float*)malloc(sizeof(float)*iAlignUp(int(locations.size()), 640));

			// Do intensity patches. Iterator over outer features vector and intensity patches.
			int i = 0;
			for ( vector<svlIntensityPatch*>::const_iterator p = dict._intEntries.begin(); p != dict._intEntries.end(); ++p ) {
				// Get pitch index of corresponding channel.
				int pix = channels2pitches[ (*p)->channel ];

				svlCodeProfiler::tic(inner_cuda);

				if ( cachedPatches.empty() ) {
					// Calculate patch features and pad patch.
					float patchSum = 0.0;
					float patchNorm = 0.0;
					unsigned char *pp = (unsigned char *)(*p)->patch->imageData;
					for ( int y=0; y<(*p)->patch->height; ++y ) {
						for ( int x=0; x<(*p)->patch->width; ++x ) {
							patchSum += (float)pp[x];
							patchNorm += (float)(pp[x] * pp[x]);
						}
						pp += (*p)->patch->widthStep;
					}
					float N = (float)((*p)->patch->width * (*p)->patch->height);
					float patchMean = patchSum / N;
					patchNorm -= patchSum * patchMean;

					// Create response with Cuda.
					svlCudaConvolution::normCCorr_v2(_IMGS[pix], (*p)->patch, _CONV, scaled_w, scaled_h, patchMean, sqrt(patchNorm/N));

				} else {
					// TODO paulb: Cache patch padding and statistics operations.
					//cout << "Got cached patches, you know.\n";

					// Create response with Cuda.
					CachedPatch *cp = &cachedPatches[i];
#ifndef ROW_INTEGS
					svlCudaConvolution::normCCorr_v2Cached(_IMGS[pix], _CONV, cp->patch->data, cp->mask->data, cp->w, cp->h, cp->N, cp->mean, cp->std, scaled_w, scaled_h);
#else
					// Window-normalize the row integrals to _TEMP1 and _TEMP2.
					//printf("w,h: %d,%d; %w: %d; dim; %d\n", scaled_w, scaled_h, (*p)->patch->width, dim);
					/*cm->pitchSet(_TEMP1, 0);
					cm->pitchSet(_TEMP2, 0);*/
					svlCudaConvolution::windowNormalizeRow(_INTEGS[pix], _TEMP1, (*p)->patch->width, scaled_w, scaled_h, dim);
					svlCudaConvolution::windowNormalizeRow(_INTEGS2[pix], _TEMP2, (*p)->patch->width, scaled_w, scaled_h, dim);
#if HIGHLANDER_DEBUG && 0
					sprintf(c,"temp1_%d.out",xena_inner);
					cm->pitch2file(c, _TEMP1);
					sprintf(c,"temp2_%d.out",xena_inner);
					cm->pitch2file(c, _TEMP2);
#endif
					//svlCudaConvolution::normCCorr_v3(_IMGS[pix], (*p)->patch, _CONV, _TEMP1, _TEMP2, scaled_w, scaled_h, cp->mean, cp->std);
					svlCudaConvolution::normCCorr_v3Cached(_IMGS[pix], _CONV, cp->patch->data, cp->mask->data, _TEMP1, _TEMP2,
						cp->w, cp->h, cp->N, cp->mean, cp->std, scaled_w, scaled_h);
#endif
				}
				
#if HIGHLANDER_DEBUG
				sprintf(c,"resp1_%d.out",xena_inner);
				cm->pitch2file(c, _CONV);
#endif

#if 0
				// Calculate response with OpenCV -- Keep this around for fast debugging.
				IplImage *ipl = cvCreateImage( cvSize(scaled_w,scaled_h), IPL_DEPTH_8U, 1);
				cm->IplFromGPU<unsigned char>(ipl, _IMGS[pix]);
				IplImage *integi = cvCreateImage( cvSize(scaled_w+1,scaled_h+1), IPL_DEPTH_32S, 1);
				IplImage *integi2 = cvCreateImage( cvSize(scaled_w+1,scaled_h+1), IPL_DEPTH_64F, 1);
				cvIntegral(ipl, integi, integi2);
				/*ipl2FloatFile<int>("cv_integ.out", integi);
				ipl2FloatFile<double>("cv_integ2.out", integi2);*/
				// Do convolution and window normalization.
				IplImage *response = cvCreateImage(cvSize(ipl->width - (*p)->patch->width + 1,
					ipl->height - (*p)->patch->height + 1), IPL_DEPTH_32F, 1);
				IplImage *convolution_response = cvCreateImage(cvSize(ipl->width - (*p)->patch->width + 1,
					ipl->height - (*p)->patch->height + 1), IPL_DEPTH_32F, 1);
				cvMatchTemplate(ipl, (*p)->patch, convolution_response, CV_TM_CCORR);
				IplImage *patch = (*p)->patch;
				for (int y = 0; y < ipl->height - patch->height + 1; ++y ) {
					const unsigned char *p = &CV_IMAGE_ELEM(ipl, unsigned char, y, 0);
					const int *ps = &CV_IMAGE_ELEM(integi, int, y, 0);
					const int *ps2 = &CV_IMAGE_ELEM(integi, int, y + patch->height, 0);
					const double *psSq = &CV_IMAGE_ELEM(integi2, double, y, 0);
					const double *psSq2 = &CV_IMAGE_ELEM(integi2, double, y + patch->height, 0);
					float *q = &CV_IMAGE_ELEM(response, float, y, 0);
					for (int x = 0; x < ipl->width - patch->width + 1; x++, p++) {
						float windowSum = (float)(ps[x] + ps2[x + patch->width] - 
							ps[x + patch->width] - ps2[x]);
						float windowNorm = (float)(psSq[x] + psSq2[x + patch->width] - 
							psSq[x + patch->width] - psSq2[x]);

						float convolutionSum = CV_IMAGE_ELEM(convolution_response, float, y, x);
						windowNorm -= windowSum * windowSum / N;
						*q = (float)convolutionSum - patchMean * windowSum;
						*q /= sqrt(patchNorm * windowNorm);
						++q;
					}
				}
				cvReleaseImage(&convolution_response);
				cm->IplToGPU<float>(response, _CONV);
#if HIGHLANDER_DEBUG
				sprintf(c,"resp_gold2_%d.out",xena_inner);
				printf(" Out'ing %s\n", c);
				//cm->pitch2file(c, _CONV);
				iplimage2file<float>(c, response);
				cvReleaseImage(&response);
#endif
#endif

				// Max operation: _CONV => _TEMP1 (becomes response image).
				svlCudaMath::maxWindow(_CONV, _TEMP1, (*p)->rect.width, (*p)->rect.height, scaled_w, scaled_h, cudagrid);
				svlCodeProfiler::toc(inner_cuda);

				// Get out features and place in "features."
				// TODO paulb: Move all of this onto the GPU: compute locations there, but also keep them and use them there.
				svlCodeProfiler::tic(read_features);
				// Use device-side lookup to stack address into linear memory for faster copying out.
				// TODO: Move locations to the GPU right away ... wait ... Yes! Do that, and then this stuff on the device.

				if ( !use_cuda_dtree ) {
					// Create true indexing locations by adding rect offset.
					for ( vector<CvPoint>::iterator loc=locations.begin(); loc!=locations.end(); ++loc ) {
						loc->x += (*p)->rect.x;
						loc->y += (*p)->rect.y;
					}

					// Fill index image.
					const int max_row = cm->fillIndexImage(locations, ixImage);
					// Copy indices over and perform lookup. TEMP1 for data; CONV for addresses, TEMP2 for data out; results end up in "results."
					cm->linearRandomAccess(ixImage, max_row, results, _TEMP1, _CONV, _TEMP2);
					// Undo index offset for next patch.

					// Place features in vectors.
					float *ptr = results;
					const int i_off = i + outputOffset;
					for ( vector<vector<double> >::iterator ot = features.begin(); ot != features.end(); ++ot ) {
						(*ot)[ i_off ] = *ptr++;
					}

					for ( vector<CvPoint>::iterator loc=locations.begin(); loc!=locations.end(); ++loc ) {
						loc->x -= (*p)->rect.x;
						loc->y -= (*p)->rect.y;
					}
				} else {
					/*// Create true indexing locations by adding rect offset.
					for ( vector<CvPoint>::iterator loc=locations.begin(); loc!=locations.end(); ++loc ) {
						loc->x += (*p)->rect.x;
						loc->y += (*p)->rect.y;
					}
					// Add rect offset to locline; save to ixline.

					// Fill an index line instead.
					//svlCudaMemHandler::fillIndexLine(locations, ixline);
					svlCudaMemHandler::CvPoint2IntLine(locations, ixline, true);*/

					//cm->line2file("ixline.out", ixline);
					CvPoint pt;
					pt.x = (*p)->rect.x;
					pt.y = (*p)->rect.y;
					//printf("add x,y: %d,%d\n", pt.x, pt.y);
					//cm->line2file("locline_pre.out", locline);
					svlCudaMemHandler::addCvPoint2Line(locline, pt, ixline);
					//cm->line2file("locline.out", locline);
					//exit(-1);

					//cm->line2file("ixline.out", ixline);
					svlCudaMemHandler::linear2DRandomAccess(ixline, _TEMP1, dfeatures, i, feature_stride, int(locations.size()));
					//cm->line2file("dfeat.out", dfeatures);

					/*for ( vector<CvPoint>::iterator loc=locations.begin(); loc!=locations.end(); ++loc ) {
						loc->x -= (*p)->rect.x;
						loc->y -= (*p)->rect.y;
					}*/
				}

				svlCodeProfiler::toc(read_features);

				++i;

	#if HIGHLANDER_DEBUG
				++xena_inner;
	#endif
			}
			free(results);

			//cm->line2file("dfeat.out", dfeatures);

			//printf("num feat: %d\n", (int)dict._intEntries.size());

	#if HIGHLANDER_DEBUG
			sprintf(c,"allfeat_%d.out",xena_outer);
			FILE *ffeat = fopen(c,"wb");
			for ( int i=0; i<(int)features.size(); ++i ) {
				for ( int j=0; j<(int)features[i].size(); ++j ) {
					fwrite(&features[i][j],sizeof(double),1,ffeat);
				}
			}
			fclose(ffeat);
	#endif

			// TODO paulb: Restore this.
			/*// Free all memory from the normalized patches
			for (unsigned j = 0; j < svlPatchDictionary::_normalizedDepthPatches.size();j++) 
			{
				cvReleaseImage(&svlPatchDictionary::_normalizedDepthPatches[j]);
			}
			svlPatchDictionary::_normalizedDepthPatches.clear();

			// Free all memory from response images
			clearAbsResponses();*/

			svlCodeProfiler::toc(extract_handler);

			/*if (DEBUG_OUTPUT) {
				string filename = string("svlSlidingWindowDetector.classifyImage.")
					+ toString(scale) + string(".txt");
				ofstream ofs(filename.c_str(), ios::app);
				for (unsigned i = 0; i < features.size(); i++) {
					ofs << (scale * locations[i].x) << " " << (scale * locations[i].y) << " "
						<< (scale * _windowWidth) << " " << (scale * _windowHeight);
					for (unsigned j = 0; j < features[i].size(); j++) {
						ofs << " " << features[i][j];
					}
					ofs << "\n";
				}
				ofs.close();
			}*/

			SVL_LOG(SVL_LOG_VERBOSE, "...done");
			SVL_LOG(SVL_LOG_VERBOSE, "Running multi-channel classifier on " << locations.size() << " windows...");

			// Evaluate model.
			svlCodeProfiler::tic(eval_model);

#if HIGHLANDER_DEBUG
			sprintf(c,"batch_%d.out",xena_outer);
			FILE *ffeat2 = fopen(c,"wb");
#endif
			if ( !use_cuda_dtree ) {

				int obj_ct = 0;
				for (unsigned i = 0; i < locations.size(); i++) {
					if ( !isSampleValid(features[i]) ) {
						continue;
					}
					float score = evaluateModel(features[i]);
#if HIGHLANDER_DEBUG
					/*for ( int j=0; j<(int)features[i].size(); ++j )
						fwrite(&features[i][j],sizeof(double),1,fo);*/
					fwrite(&score,sizeof(float),1,ffeat2);
#endif
					if (score > scoreThreshold) {
						++obj_ct;
						objects.push_back(svlObject2d(scale * locations[i].x, scale * locations[i].y,
							scale * _windowWidth, scale * _windowHeight));
						objects.back().name = _objectName;
						objects.back().pr = scoreToProbability(score);
					}
				}
				//printf("Found %d objects at scale %d,%d\n", obj_ct, scaled_w, scaled_h);
			} else {
				// Calculate decision tree nodes on device.
				dtree->evaluate(dfeatures, dict.numFeatures(), feature_stride, int(locations.size()), 100);
				//cm->line2file("res.out", dfeatures);

				// Integrate features.
				// feature_stride is correctly power-of-two-aligned. If native_pow_2 use cumsum-style operation.
				//printf("dfeats w: %d\n", dfeatures->w);
				svlCudaMath::integralLine(dfeatures, dfeatures, dfeatures->w, aligned_n_trees, native_pow_2, feature_stride);
				//cm->line2file("intres.out", dfeatures);

				// Compact results and read back out.
				svlCudaLine *res = cm->allocLine(int(locations.size()), "res");
				dtree->collectResults(dfeatures, res, int(locations.size()), feature_stride);
				float *hres = (float*)malloc(sizeof(float)*locations.size());
				cm->lineFromGPU(hres, res);

				//cm->line2file("reses.out", res);

				// Compute objectage.
				float *ptr = hres;
				int ct = 0;
				for ( vector<CvPoint>::const_iterator loc = locations.begin(); loc != locations.end(); ++loc, ++ptr ) {
					if ( isinf(*ptr) || isnan(*ptr) )
						continue;
#if HIGHLANDER_DEBUG
					fwrite(ptr,sizeof(float),1,ffeat2);
#endif
					if ( *ptr > scoreThreshold ) {
						++ct;
						//printf(" Objection!: %0.4f\n", *ptr);
						objects.push_back(svlObject2d(scale * loc->x, scale * loc->y,
							scale * _windowWidth, scale * _windowHeight));
						objects.back().name = _objectName;
						objects.back().pr = scoreToProbability(*ptr);
					}
				}
				//printf("Covered %d\n", ptr-hres);

				// Free feature memory.
				cm->freeLine(dfeatures);
				cm->freeLine(res);
				cm->freeLine(ixline);
				cm->freeLine(locline);
				free(hres);
				//if ( xena_outer == 1 ) exit(-1);
				//exit(-1);
				//use_cuda_dtree = false; // Disable further CUDA'ing.

			}
#if HIGHLANDER_DEBUG
			fclose(ffeat2);
#endif

			SVL_LOG(SVL_LOG_VERBOSE, "...done");
			svlCodeProfiler::toc(eval_model);
		}

		// Scale down each image using OpenCV, since the results have proved very temperamental with respect to that.
		svlCodeProfiler::tic(scale_handle);
		scaled_w = int( float(scaled_w) / _delta_scale );
		scaled_h = int( float(scaled_h) / _delta_scale );
		for ( int i=0; i<int(images.size()); ++i ) {
			IplImage *scaled = cvCreateImage( cvSize(scaled_w, scaled_h), images[i]->depth, images[i]->nChannels);
			cvResize(images[i], scaled);
			cm->IplToGPU<unsigned char>(scaled, _IMGS[i]);
		}
		scale *= _delta_scale;
		svlCodeProfiler::toc(scale_handle);
		//printf("exit scale: %0.4f\n", scale);

#if HIGHLANDER_DEBUG
	++xena_outer;
#endif
	}

	svlCodeProfiler::toc(prof00);

	return objects;
}

//svlObject2dFrame svlCudaSlidingWindowDetector::cudaClassifyImage(const vector<IplImage *>& images, int outputOffset)
//{
//	// Set globals and allocate temporaries.
//	//use_integral_images = false; // Set use of integral images.
//	svlCudaCoveringGrid *cudagrid = NULL; // Set no covering grid available.
//
//	// Assert that we have not been given any NULL images or images of different size.
//	assert(images[0] != NULL);
//	for ( unsigned i=1; i<images.size(); ++i ) {
//		assert(images[i] != NULL);
//		assert(images[i]->width == images[0]->width);
//		assert(images[i]->height == images[0]->height);
//	}
//
//	// Code profiling handles.
//	const int prof00 = svlCodeProfiler::getHandle("cSWD::classify");
//	const int extract_handler = svlCodeProfiler::getHandle("cSWD::extract");
//	const int read_features = svlCodeProfiler::getHandle("cSWD::read features");
//	const int inner_cuda = svlCodeProfiler::getHandle("cSWD::inner cuda");
//	const int feat_upkeep = svlCodeProfiler::getHandle("cSWD::feat upkeep");
//	const int win_locs = svlCodeProfiler::getHandle("cSWD::window locs");
//	const int eval_model = svlCodeProfiler::getHandle("cSWD::eval model");
//	const int scale_handle = svlCodeProfiler::getHandle("cSWD::scale image");
//
//	svlCodeProfiler::tic(prof00);
//
//	// Find out how many images we will need in this batch.
//	map<int,int> channels2pitches; // This will map from channels selected to device pitch indices.
//	// Loop through dictionary to see which channels will be used.
//	int gi = 0;
//	for ( vector<svlIntensityPatch*>::const_iterator p = dict._intEntries.begin(); p != dict._intEntries.end(); ++p ) {
//		if ( channels2pitches.count( (*p)->channel ) == 0 ) {
//			channels2pitches[(*p)->channel] = gi++;
//		}
//	}
//	// Copy all images over to GPU.
//	for ( int i=0; i<int(images.size()); ++i ) {
//		if ( images[i]->depth == IPL_DEPTH_8U ) {
//			cout << "Copying uchar Ipl\n";
//			cm->IplToGPU<unsigned char>( images[i], _IMGS[ channels2pitches[i] ] );
//		}
//	}
//
//#define HIGHLANDER_DEBUG 1
//#if HIGHLANDER_DEBUG
//	static int xena_outer, xena_inner; char c[32];
//#endif
//
//	// Objects to return.
//	svlObject2dFrame objects;
//	float scoreThreshold = probabilityToScore((float)_threshold);
//	int numChannels = (int)images.size();
//	double scale = 1.0;
//
//	SVL_LOG(SVL_LOG_MESSAGE, "Running classifier on " << numChannels << "-channel image");
//
//	// Create int32 image for storing indices.
//	IplImage* ixImage = cvCreateImage(cvSize(images[0]->width, images[0]->height), IPL_DEPTH_32S, 1);
//	// Features that they are not allocated afresh every time (soon to be optimized away since all feature calculation will be on the device.
//	Tsamples features;
//
//	//vector<IplImage*> scaleds;
//
//	// Set current working image scale.
//	int scaled_w = images[0]->width, scaled_h = images[0]->height;
//	// Iterate over all scales
//	while ( scaled_w >= _windowWidth && scaled_h >= _windowHeight )
//	{
//#if HIGHLANDER_DEBUG
//		// Output images in use.
//		for ( int i=0; i<int(images.size()); ++i ) {
//			sprintf(c,"img_%d_%d.out",xena_outer,i);
//			cm->pitch2file(c, _IMGS[i]);
//		}
//#endif
//
//		// Get interesting locations.
//		svlCodeProfiler::tic(win_locs);
//		vector<CvPoint> locations;
//		createInterestingWindowLocations(_IMGS[0], scaled_w, scaled_h, locations); // Get locations from channel 0.
//		svlCodeProfiler::toc(win_locs);
//
//#if HIGHLANDER_DEBUG
//		// Output all locations.
//		sprintf(c,"locs_%d.out",xena_outer);
//		FILE *floc = fopen(c,"wb");
//		for ( int i=0; i<(int)locations.size(); ++i ) {
//			fwrite(&locations[i].x,sizeof(int),1,floc);
//			fwrite(&locations[i].y,sizeof(int),1,floc);
//		}
//		fclose(floc);
//#endif
//
//		// Compute integral images for new scale.
//		for ( int i=0; i<int(images.size()); ++i ) {
//			svlCudaMath::integralBlock(_IMGS[i], _INTEGS[i], _TEMP1, _INTEGS2[i], _TEMP2, scaled_w, scaled_h);
//		}
//#if HIGHLANDER_DEBUG
//		// Output integral images.
//		for ( int i=0; i<int(images.size()); ++i ) {
//			sprintf(c,"integ_%d_%d.out",xena_outer,i);
//			cm->pitch2file(c, _IMGS[i]);
//		}
//#endif
//
//#if 0
//		// Use OpenCV for integral images. -- Apparently doesn't work.
//		cout << "Open cv'erized.\n";
//		IplImage *img = cvCreateImage( cvSize(scaled_w,scaled_h), IPL_DEPTH_8U, 1);
//		IplImage *integ = cvCreateImage( cvSize(scaled_w+1,scaled_h+1), IPL_DEPTH_32S, 1);
//		IplImage *integ2 = cvCreateImage( cvSize(scaled_w+1,scaled_h+1), IPL_DEPTH_64F, 1);
//		for ( int i=0; i<int(images.size()); ++i ) {
//			cm->IplFromGPU<unsigned char>(img, _IMGS[i]);
//			cvIntegral(img, integ, integ2);
//			cm->IplToGPU<int>(integ, _INTEGS[i]);
//			cm->IplToGPU<double>(integ2, _INTEGS2[i]);
//		}
//		cvReleaseImage(&img);
//		cvReleaseImage(&integ);
//		cvReleaseImage(&integ2);
//#endif
//
//		// Compute features.
//		/*SVL_LOG(SVL_LOG_VERBOSE, "Computing features over " << locations.size() << " windows for image size "
//			<< scaledImages[0]->width << " x " << scaledImages[0]->height << "...");*/
//		svlCodeProfiler::tic(extract_handler);
//		svlCodeProfiler::tic(feat_upkeep);
//
//		// Allocate memory for return vectors.
//		if ( features.size() != locations.size() )
//			features.resize(locations.size());
//		// Resize features's inner vectors to needed size.
//		unsigned neededSize = outputOffset + unsigned(dict.numFeatures());
//		for ( vector<vector<double> >::iterator it = features.begin(); it != features.end(); ++it ) {
//			if ( it->size() < neededSize)
//				it->resize(neededSize);
//		}
//
//		svlCodeProfiler::toc(feat_upkeep);
//
//		//svlDepthPatchDefinition::_locationsize = _locationsize;
//
//		// TODO paulb: Put this all back in later.
//		//Check if any channels are depth or abs
//		/*vector<bool> isDepthChannel(images.size(),false);
//		vector<bool> isAbsChannel(images.size(), false);*/
//		/*for (unsigned i = 0; i < _entries.size(); i++)
//		{
//			if (_entries[i]->type == SVL_DEPTH_PATCH)
//				isDepthChannel[ _entries[i]->channel ] = true;
//
//			if (_entries[i]->type == SVL_ABS_PATCH)
//				isAbsChannel[ _entries[i]->channel ] = true;
//		}
//		for (unsigned i = 0; i < images.size(); i++)
//		{
//			if (isDepthChannel[i])
//				normalizeDepthPatches(images[i], locations);
//
//			if (isAbsChannel[i])
//				computeAbsResponses(images[i]);
//		}*/
//
//		// Compute feature responses from dictionary.
//		// TODO paulb: Convert these to linear memory on the GPU.
//		float *results = (float*)malloc(sizeof(float)*iAlignUp(int(locations.size()), 640));
//
//		// Do intensity patches. Iterator over outer features vector and intensity patches.
//		int i = 0;
//		for ( vector<svlIntensityPatch*>::const_iterator p = dict._intEntries.begin(); p != dict._intEntries.end(); ++p ) {
//			// Get pitch index of corresponding channel.
//			int pix = channels2pitches[ (*p)->channel ];
//			printf("channel %d\n", (*p)->channel);
//
//			// TODO paulb: Cache patch padding and statistics operations.
//			// Calculate patch features and pad patch.
//			float patchSum = 0.0;
//			float patchNorm = 0.0;
//			unsigned char *pp = (unsigned char *)(*p)->patch->imageData;
//			for ( int y=0; y<(*p)->patch->height; ++y ) {
//				for ( int x=0; x<(*p)->patch->width; ++x ) {
//					patchSum += (float)pp[x];
//					patchNorm += (float)(pp[x] * pp[x]);
//				}
//				pp += (*p)->patch->widthStep;
//			}
//			float N = (float)((*p)->patch->width * (*p)->patch->height);
//			float patchMean = patchSum / N;
//			patchNorm -= patchSum * patchMean;
//
//			svlCodeProfiler::tic(inner_cuda);
//#define USE_CUDA_INNER 0
//#if USE_CUDA_INNER
//			// Create response with Cuda.
//#if 0
//			// Normalized cross correlation via convolution and normalization.
//			svlCudaConvolution::conv2d_v2<unsigned char>(_IMGS[pix], (*p)->patch, _CONV, scaled_w, scaled_h);
//#if HIGHLANDER_DEBUG
//			sprintf(c,"conv_%d.out",xena_inner);
//			cm->pitch2file(c, _CONV);
//#endif
//			svlCudaConvolution::windowNormalize1DimSub(_CONV, _INTEGS[pix], _INTEGS2[pix],
//				patchMean, sqrt(patchNorm/N), (*p)->patch->width, (*p)->patch->height, scaled_w, scaled_h);
//#else
//			svlCudaConvolution::normCCorr(_IMGS[pix], (*p)->patch, _CONV, scaled_w, scaled_h, patchMean, sqrt(patchNorm/N));
//#endif
//#if HIGHLANDER_DEBUG
//			sprintf(c,"resp_%d.out",xena_inner);
//			cm->pitch2file(c, _CONV);
//#endif
//#else
//			// Response with OpenCV.
//			IplImage *ipl = cvCreateImage( cvSize(scaled_w,scaled_h), IPL_DEPTH_8U, 1);
//			cm->IplFromGPU<unsigned char>(ipl, _IMGS[pix]);
//			IplImage *integi = cvCreateImage( cvSize(scaled_w+1,scaled_h+1), IPL_DEPTH_32S, 1);
//			IplImage *integi2 = cvCreateImage( cvSize(scaled_w+1,scaled_h+1), IPL_DEPTH_64F, 1);
//			cvIntegral(ipl, integi, integi2);
//			/*ipl2FloatFile<int>("cv_integ.out", integi);
//			ipl2FloatFile<double>("cv_integ2.out", integi2);*/
//			// Do convolution and window normalization.
//			IplImage *response = cvCreateImage(cvSize(ipl->width - (*p)->patch->width + 1,
//				ipl->height - (*p)->patch->height + 1), IPL_DEPTH_32F, 1);
//			IplImage *convolution_response = cvCreateImage(cvSize(ipl->width - (*p)->patch->width + 1,
//				ipl->height - (*p)->patch->height + 1), IPL_DEPTH_32F, 1);
//			cvMatchTemplate(ipl, (*p)->patch, convolution_response, CV_TM_CCORR);
//			IplImage *patch = (*p)->patch;
//			for (int y = 0; y < ipl->height - patch->height + 1; ++y ) {
//				const unsigned char *p = &CV_IMAGE_ELEM(ipl, unsigned char, y, 0);
//				const int *ps = &CV_IMAGE_ELEM(integi, int, y, 0);
//				const int *ps2 = &CV_IMAGE_ELEM(integi, int, y + patch->height, 0);
//				const double *psSq = &CV_IMAGE_ELEM(integi2, double, y, 0);
//				const double *psSq2 = &CV_IMAGE_ELEM(integi2, double, y + patch->height, 0);
//				float *q = &CV_IMAGE_ELEM(response, float, y, 0);
//				for (int x = 0; x < ipl->width - patch->width + 1; x++, p++) {
//					float windowSum = (float)(ps[x] + ps2[x + patch->width] - 
//						ps[x + patch->width] - ps2[x]);
//					float windowNorm = (float)(psSq[x] + psSq2[x + patch->width] - 
//						psSq[x + patch->width] - psSq2[x]);
//
//					float convolutionSum = CV_IMAGE_ELEM(convolution_response, float, y, x);
//					windowNorm -= windowSum * windowSum / N;
//					*q = (float)convolutionSum - patchMean * windowSum;
//					*q /= sqrt(patchNorm * windowNorm);
//					++q;
//				}
//			}
//			cvReleaseImage(&convolution_response);
//			cm->IplToGPU<float>(response, _CONV);
//#if HIGHLANDER_DEBUG
//			sprintf(c,"resp_gold2_%d.out",xena_inner);
//			printf(" Out'ing %s\n", c);
//			//cm->pitch2file(c, _CONV);
//			iplimage2file<float>(c, response);
//			cvReleaseImage(&response);
//#endif
//#endif
//
//			// Max operation: _CONV => _TEMP1 (becomes response image).
//			svlCudaMath::maxWindow(_CONV, _TEMP1, (*p)->rect.width, (*p)->rect.height, scaled_w, scaled_h, cudagrid);
//			svlCodeProfiler::toc(inner_cuda);
//
//			// Get out features and place in "features."
//			// TODO paulb: Move all of this onto the GPU: compute locations there, but also keep them and use them there.
//			svlCodeProfiler::tic(read_features);
//			// Use device-side lookup to stack address into linear memory for faster copying out.
//			// Create true indexing locations by adding rect offset.
//			for ( vector<CvPoint>::iterator loc=locations.begin(); loc!=locations.end(); ++loc ) {
//				loc->x += (*p)->rect.x;
//				loc->y += (*p)->rect.y;
//			}
//			// Fill index image.
//			const int max_row = cm->fillIndexImage(locations, ixImage);
//			// Copy indices over and perform lookup. TEMP1 for data; CONV for addresses, TEMP2 for data out; results end up in "results."
//			cm->linearRandomAccess(ixImage, max_row, results, _TEMP1, _CONV, _TEMP2);
//			// Undo index offset for next patch.
//			for ( vector<CvPoint>::iterator loc=locations.begin(); loc!=locations.end(); ++loc ) {
//				loc->x -= (*p)->rect.x;
//				loc->y -= (*p)->rect.y;
//			}
//
//			// Place features in vectors.
//			float *ptr = results;
//			const int i_off = i + outputOffset;
//			for ( vector<vector<double> >::iterator ot = features.begin(); ot != features.end(); ++ot ) {
//				(*ot)[ i_off ] = *ptr++;
//			}
//			svlCodeProfiler::toc(read_features);
//			++i;
//
//#if HIGHLANDER_DEBUG
//			++xena_inner;
//#endif
//		}
//		free(results);
//
//		//printf("num feat: %d\n", (int)dict._intEntries.size());
//
//		/*{	static int xena; char c[32]; sprintf(c,"allfeat_%d.out",xena++);
//			FILE *fo = fopen(c,"wb");
//			for ( int i=0; i<(int)features.size(); ++i ) {
//				for ( int j=0; j<(int)features[i].size(); ++j ) {
//					fwrite(&features[i][j],sizeof(double),1,fo);
//				}
//			}
//			fclose(fo);	}*/
//
//		// TODO paulb: Restore this.
//		/*// Free all memory from the normalized patches
//		for (unsigned j = 0; j < svlPatchDictionary::_normalizedDepthPatches.size();j++) 
//		{
//			cvReleaseImage(&svlPatchDictionary::_normalizedDepthPatches[j]);
//		}
//		svlPatchDictionary::_normalizedDepthPatches.clear();
//
//		// Free all memory from response images
//		clearAbsResponses();*/
//
//		svlCodeProfiler::toc(extract_handler);
//
//		/*if (DEBUG_OUTPUT) {
//			string filename = string("svlSlidingWindowDetector.classifyImage.")
//				+ toString(scale) + string(".txt");
//			ofstream ofs(filename.c_str(), ios::app);
//			for (unsigned i = 0; i < features.size(); i++) {
//				ofs << (scale * locations[i].x) << " " << (scale * locations[i].y) << " "
//					<< (scale * _windowWidth) << " " << (scale * _windowHeight);
//				for (unsigned j = 0; j < features[i].size(); j++) {
//					ofs << " " << features[i][j];
//				}
//				ofs << "\n";
//			}
//			ofs.close();
//		}*/
//
//		SVL_LOG(SVL_LOG_VERBOSE, "...done");
//		SVL_LOG(SVL_LOG_VERBOSE, "Running multi-channel classifier on " << locations.size() << " windows...");
//
//		// Evaluate model.
//		svlCodeProfiler::tic(eval_model);
//
//#define FEAT_OUT 0
//#if FEAT_OUT
//		{	static int xena; char c[32]; sprintf(c,"batch_out_%d.out",xena++);
//			FILE *fo = fopen(c,"wb");
//#endif
//		int obj_ct = 0;
//		for (unsigned i = 0; i < locations.size(); i++) {
//			if ( !isSampleValid(features[i]) ) {
//				continue;
//			}
//			float score = evaluateModel(features[i]);
//#if FEAT_OUT
//			for ( int j=0; j<(int)features[i].size(); ++j )
//				fwrite(&features[i][j],sizeof(double),1,fo);
//			double gd = score;
//			fwrite(&gd,sizeof(double),1,fo);
//#endif
//			if (score > scoreThreshold) {
//				++obj_ct;
//				objects.push_back(svlObject2d(scale * locations[i].x, scale * locations[i].y,
//					scale * _windowWidth, scale * _windowHeight));
//				objects.back().name = _objectName;
//				objects.back().pr = scoreToProbability(score);
//			}
//		}
//		printf("Found %d objects at scale %d,%d\n", obj_ct, scaled_w, scaled_h);
//#if FEAT_OUT
//			fclose(fo);	}
//#endif
//
//		SVL_LOG(SVL_LOG_VERBOSE, "...done");
//		svlCodeProfiler::toc(eval_model);
//
//		// Scale down each image channel on the GPU.
//		svlCodeProfiler::tic(scale_handle);
//		//printf("Scaling: %0.3f\n", _delta_scale);
//		int new_scaled_w = int( float(scaled_w) / _delta_scale );
//		int new_scaled_h = int( float(scaled_h) / _delta_scale );
//		for ( int i=0; i<int(images.size()); ++i ) {
//#if 0
//			// Scale images: IMG -> TEMP1
//			svlCudaMath::resize2D(_IMGS[i], _TEMP1, 1.0f/float(_delta_scale), scaled_w, scaled_h);
//			// Zeralize IMG and replace with scales down TEMP1.
//			cm->pitchSet(_IMGS[i], 0);
//			cm->pitchToPitch(_IMGS[i], _TEMP1, new_scaled_w, new_scaled_h);
//#else
//			//IplImage *ipl = cvCreateImage( cvSize(scaled_w, scaled_h), IPL_DEPTH_8U, 1);
//			//cm->IplFromGPU<unsigned char>(ipl, _IMGS[i]);
//			IplImage *scaled = cvCreateImage( cvSize(new_scaled_w, new_scaled_h), images[i]->depth, images[i]->nChannels);
//			cvResize(images[i], scaled);
//			cm->IplToGPU<unsigned char>(scaled, _IMGS[i]);
//#endif
//		}
//		scale *= _delta_scale;
//		scaled_w = new_scaled_w; // Decremented scales.
//		scaled_h = new_scaled_h;
//		svlCodeProfiler::toc(scale_handle);
//		printf("exit scale: %0.4f\n", scale);
//
//#if HIGHLANDER_DEBUG
//	++xena_outer;
//#endif
//	}
//
//	svlCodeProfiler::toc(prof00);
//
//	return objects;
//}

#else
// Classifies a multi-channel image.
svlObject2dFrame svlCudaSlidingWindowDetector::cudaClassifyImage(const vector<IplImage *>& images, int outputOffset)
{
	//assert(!images.empty());

	// Set globals and allocate temporaries.
	use_integral_images = true; // Set use of integral images.
	svlCudaCoveringGrid *cudagrid = NULL; // Set no covering grid available.
	//int gi;
	IplImage *img; // An image pointer.

	int prof00 = svlCodeProfiler::getHandle("cSWD::classify");
	svlCodeProfiler::tic(prof00);
	int integ_clock = svlCodeProfiler::getHandle("cSWD::integral images");
	const int extract_handler = svlCodeProfiler::getHandle("cSWD::extract");
	const int read_features = svlCodeProfiler::getHandle("cSWD::read features");
	const int inner_cuda = svlCodeProfiler::getHandle("cSWD::inner cuda");
	const int feat_upkeep = svlCodeProfiler::getHandle("cSWD::feat upkeep");

	// Assert that we have not been given any NULL images or images of different size.
	for ( unsigned i=0; i<images.size(); ++i ) {
		assert(images[i] != NULL);
		assert(images[i]->width == images[0]->width);
		assert(images[i]->height == images[0]->height);
	}

	svlObject2dFrame objects; // Return objects.
	double scale = 1.0;
	float scoreThreshold = probabilityToScore((float)_threshold);
	int numChannels = (int)images.size();

	SVL_LOG(SVL_LOG_MESSAGE, "Running classifier on " << numChannels << "-channel image");

	// Copy of original images for color conversion and rescaling
	vector<IplImage*> scaledImages(numChannels), intImages(numChannels), int2Images(numChannels);
	for ( int i=0; i<numChannels; ++i ) {
		scaledImages[i] = cvCloneImage(images[i]);
		if ( !use_integral_images ) {
			intImages[i] = NULL;
			int2Images[i] = NULL;
		} else {
			svlCodeProfiler::tic(integ_clock);
			intImages[i] = cvCreateImage(cvSize(scaledImages[i]->width + 1, scaledImages[i]->height + 1), IPL_DEPTH_32S, scaledImages[i]->nChannels);
			int2Images[i] = cvCreateImage(cvSize(scaledImages[i]->width + 1, scaledImages[i]->height + 1), IPL_DEPTH_64F, scaledImages[i]->nChannels);
			cvIntegral(scaledImages[i], intImages[i], int2Images[i]);
			svlCodeProfiler::toc(integ_clock);
			// Try doing integral image in a quad.
			/*cvSetImageROI(scaledImages[i], cvRect(0,0, scaledImages[i]->width/2, scaledImages[i]->height/2));
			IplImage *icka = cvCreateImage(cvSize(scaledImages[i]->width + 1, scaledImages[i]->height + 1), IPL_DEPTH_32S, scaledImages[i]->nChannels);
			cvSetImageROI(icka, cvRect(0,0, scaledImages[i]->width/2 + 1, scaledImages[i]->height/2 + 1));
			cvIntegral(scaledImages[i], icka);
			iplimage2file<int>("icka.out", icka);
			exit(-1);*/
		}
	}
	IplImage* responseImage = cvCreateImage(cvSize(scaledImages[0]->width, scaledImages[0]->height), IPL_DEPTH_32F, 1); // Create float image for storing response.
	IplImage* ixImage = cvCreateImage(cvSize(images[0]->width, images[0]->height), IPL_DEPTH_32S, 1); // Create int32 image for storing indices.

	Tsamples features; // So that they are not allocated afresh every time.

	// Iterate over all scales
	while ((scaledImages[0]->width >= _windowWidth) && (scaledImages[0]->height >= _windowHeight)) {
		int prof0 = svlCodeProfiler::getHandle("cSWD::window locs");
		svlCodeProfiler::tic(prof0);

		/*vector<CvPoint> locations;
		createWindowLocations(scaledImages[0]->width, scaledImages[0]->height, locations);
		removeUninterestingLocations(scaledImages[0], locations);*/

		// Get interesting locations, sped up with Cuda.
		vector<CvPoint> locations;
		createInterestingWindowLocations(scaledImages[0], locations);

		//createInterestingWindowLocations(scaledImages[0], locations, cudamem);
		//svlCudaCoveringGrid cudagrid(scaledImages[0]->width, scaledImages[0]->height, 16, 16);
		//cudagrid.clearGrid();
		//computeCudaCoveringGrid(locations, 16, 16, 16, 16, cudagrid); // Mark spots 16 away so as not to miss and intermediate blocks.
		//computeCudaCoveringGrid(locations, 16, 16, 32, 32, cudagrid); // Mark spots 16 away so as not to miss and intermediate blocks.

		svlCodeProfiler::toc(prof0);

		// Run classifier
		SVL_LOG(SVL_LOG_VERBOSE, "Computing features over " << locations.size() << " windows for image size "
			<< scaledImages[0]->width << " x " << scaledImages[0]->height << "...");

		// Compute features.
		//extract(locations, scaledImages, intImages, int2Images, responseImage, ixImage, features);
		
		svlCodeProfiler::tic(extract_handler);
		svlCodeProfiler::tic(feat_upkeep);

		// Allocate memory for return vectors.
		if ( features.size() != locations.size() )
			features.resize(locations.size());
		// Resize features's inner vectors to needed size.
		unsigned neededSize = outputOffset + unsigned(dict.numFeatures());
		for ( vector<vector<double> >::iterator it = features.begin(); it != features.end(); ++it ) {
			if ( it->size() < neededSize)
				it->resize(neededSize);
		}

		svlCodeProfiler::toc(feat_upkeep);

		//int numChannels = (int)images.size();

		//svlDepthPatchDefinition::_locationsize = _locationsize;

		// TODO paulb: Put this all back in later.
		//Check if any channels are depth or abs
		/*vector<bool> isDepthChannel(images.size(),false);
		vector<bool> isAbsChannel(images.size(), false);*/
		/*for (unsigned i = 0; i < _entries.size(); i++)
		{
			if (_entries[i]->type == SVL_DEPTH_PATCH)
				isDepthChannel[ _entries[i]->channel ] = true;

			if (_entries[i]->type == SVL_ABS_PATCH)
				isAbsChannel[ _entries[i]->channel ] = true;
		}
		for (unsigned i = 0; i < images.size(); i++)
		{
			if (isDepthChannel[i])
				normalizeDepthPatches(images[i], locations);

			if (isAbsChannel[i])
				computeAbsResponses(images[i]);
		}*/


		// Compute feature responses from dictionary.


		float *results = (float*)malloc(sizeof(float)*iAlignUp(int(locations.size()), 640));

		// Do intensity patches. Iterator over outer features vector and intensity patches.
		int i = 0;
		for ( vector<svlIntensityPatch*>::const_iterator p = dict._intEntries.begin(); p != dict._intEntries.end(); ++p ) {
			// Get image and copy to GPU at pitch 0.
			img = scaledImages[(*p)->channel];
			/*printf("  patch %d\n", i);
			printf("   img w,h: %d,%d\n", img->width, img->height);*/

			// Copy image to GPU.
			if ( !cm->pitchHasCached((void*)img, _IMG) ) {
				cm->IplToGPU<unsigned char>(img, _IMG);
			}
#ifdef DEVICE_INTEGRALS
			// Check if this image has been integralized on the GPU.
			if ( !cm->pitchHasCached((void*)img, _INTEG) ) {
				// Calculate new integral images.
				svlCudaMath::integral(_IMG, _INTEG, _INTEG2, img->width, img->height);
				// Debug check output.
				/*cm->pitch2file("img.out", _IMG);
				cm->pitch2file("integ.out", _INTEG);
				cm->pitch2file("integ2.out", _INTEG2);*/
				// Set "sources" of integral images to reflect base image.
				cm->pitchSetSource((void*)img, _INTEG);
				cm->pitchSetSource((void*)img, _INTEG2);
			}
#else
			// Copy integral images to GPU.
			if ( use_integral_images
			  && intImages[ (*p)->channel ] != NULL
			  && int2Images[ (*p)->channel ] != NULL 
			  && !cm->pitchHasCached((void*)intImages[ (*p)->channel ], _INTEG) ) {
				cm->IplToGPU<int>(intImages[ (*p)->channel ], _INTEG);
				cm->IplToGPU<double>(int2Images[ (*p)->channel ], _INTEG2);
			}
#endif

			// Calculate patch features and pad patch.
			float patchSum = 0.0;
			float patchNorm = 0.0;
			unsigned char *pp = (unsigned char *)(*p)->patch->imageData;
			for ( int y=0; y<(*p)->patch->height; ++y ) {
				for ( int x=0; x<(*p)->patch->width; ++x ) {
					patchSum += (float)pp[x];
					patchNorm += (float)(pp[x] * pp[x]);
				}
				pp += (*p)->patch->widthStep;
			}
			float N = (float)((*p)->patch->width * (*p)->patch->height);
			float patchMean = patchSum / N;
			patchNorm -= patchSum * patchMean;

			svlCodeProfiler::tic(inner_cuda);

			if ( !use_integral_images ) {
				// Normalized cross correlation via convolution and normalization.
				svlCudaConvolution::conv2d_v2<unsigned char>(_IMG, (*p)->patch, _CONV, img->width, img->height);
#ifndef DEVICE_INTEGRALS
				svlCudaConvolution::windowNormalize1(_CONV, _INTEG, _INTEG2,
					patchMean, sqrt(patchNorm/N), (*p)->patch->width, (*p)->patch->height, img->width, img->height);
#else
				svlCudaConvolution::windowNormalize1DimSub(_CONV, _INTEG, _INTEG2,
					patchMean, sqrt(patchNorm/N), (*p)->patch->width, (*p)->patch->height, img->width, img->height);
#endif
			} else {
				// Let's have a debug outing.
				svlCudaConvolution::conv2d_v2<unsigned char>(_IMG, (*p)->patch, _CONV, img->width, img->height);
				svlCudaConvolution::windowNormalize1(_CONV, _INTEG, _INTEG2,
					patchMean, sqrt(patchNorm/N), (*p)->patch->width, (*p)->patch->height, img->width, img->height);
			}
			//cm->pitch2file("resp.out", _CONV); exit(-1);

			// Max operation.
			svlCudaMath::maxWindow(_CONV, _TEMP1, (*p)->rect.width, (*p)->rect.height, img->width, img->height, cudagrid);

#ifndef RAND_ACCESS
			// Get results back directly.
			cm->floatIplFromGPU(responseImage, _TEMP1);

			/*cm->pitch2file("conv.out", _CONV);
			cm->pitch2file("resp.out", _TEMP1);
			exit(-1);*/
			svlCodeProfiler::toc(inner_cuda);

			svlCodeProfiler::tic(read_features);

			// Read out features.
			vector<CvPoint>::const_iterator w = locations.begin();
			vector<vector<double> >::iterator ot = features.begin();
			while ( w != locations.end() ) {
				(*ot)[ i + outputOffset ] = CV_IMAGE_ELEM(responseImage, float, (*p)->rect.y + w->y, (*p)->rect.x + w->x);
				++w; ++ot;
			}
			/*FILE *fo = fopen("features_gold.out","wb"); float gf;
			while ( w != locations.end() ) {
				(*ot)[ i + featuresOffset ] = gf = CV_IMAGE_ELEM(response, float, (*p)->rect.y + w->y, (*p)->rect.x + w->x);
				fwrite(&gf,sizeof(float),1,fo);
				++w; ++ot;
			}
			fclose(fo);
			exit(-1);*/

			svlCodeProfiler::toc(read_features);
#else
			// Use device-side lookup to stack address into linear stretch of memory.
			// Create true indexing locations by adding rect offset.
			for ( vector<CvPoint>::iterator loc=locations.begin(); loc!=locations.end(); ++loc ) {
				loc->x += (*p)->rect.x;
				loc->y += (*p)->rect.y;
			}
			// Fill index image.
			const int max_row = cm->fillIndexImage(locations, ixImage);

			// Copy indices over and perform lookup. TEMP1 for data; CONV for addresses, TEMP2 for data out; results in "results."
			cm->linearRandomAccess(ixImage, max_row, results, _TEMP1, _CONV, _TEMP2);

			// Undo index offset for next patch.
			for ( vector<CvPoint>::iterator loc=locations.begin(); loc!=locations.end(); ++loc ) {
				loc->x -= (*p)->rect.x;
				loc->y -= (*p)->rect.y;
			}
			svlCodeProfiler::toc(inner_cuda);

			svlCodeProfiler::tic(read_features);

			// Place features in vectors.
			float *ptr = results;
			const int i_off = i + outputOffset;
			for ( vector<vector<double> >::iterator ot = features.begin(); ot != features.end(); ++ot ) {
				(*ot)[ i_off ] = *ptr++;
			}

			svlCodeProfiler::toc(read_features);
#endif
			++i;
		}
		free(results);

		// TODO paulb: Restore this.
		/*// Free all memory from the normalized patches
		for (unsigned j = 0; j < svlPatchDictionary::_normalizedDepthPatches.size();j++) 
		{
			cvReleaseImage(&svlPatchDictionary::_normalizedDepthPatches[j]);
		}
		svlPatchDictionary::_normalizedDepthPatches.clear();

		// Free all memory from response images
		clearAbsResponses();*/

		svlCodeProfiler::toc(extract_handler);


		if (DEBUG_OUTPUT) {
			string filename = string("svlSlidingWindowDetector.classifyImage.")
				+ toString(scale) + string(".txt");
			ofstream ofs(filename.c_str(), ios::app);
			for (unsigned i = 0; i < features.size(); i++) {
				ofs << (scale * locations[i].x) << " " << (scale * locations[i].y) << " "
					<< (scale * _windowWidth) << " " << (scale * _windowHeight);
				for (unsigned j = 0; j < features[i].size(); j++) {
					ofs << " " << features[i][j];
				}
				ofs << "\n";
			}
			ofs.close();
		}

		SVL_LOG(SVL_LOG_VERBOSE, "...done");
		SVL_LOG(SVL_LOG_VERBOSE, "Running multi-channel classifier on " << locations.size() << " windows...");

		int prof2 = svlCodeProfiler::getHandle("cSWD::eval model");
		svlCodeProfiler::tic(prof2);

		for (unsigned i = 0; i < locations.size(); i++) {
			if(!isSampleValid(features[i])) continue;
			float score = evaluateModel(features[i]);
			if (score > scoreThreshold) {
				objects.push_back(svlObject2d(scale * locations[i].x, scale * locations[i].y,
					scale * _windowWidth, scale * _windowHeight));
				objects.back().name = _objectName;
				objects.back().pr = scoreToProbability(score);
			}
		}

		SVL_LOG(SVL_LOG_VERBOSE, "...done");

		svlCodeProfiler::toc(prof2);

		int prof3 = svlCodeProfiler::getHandle("cSWD::scale image");
		svlCodeProfiler::tic(prof3);

		// Scale down each image channel
		scale *= _delta_scale;
		for (int i = 0; i < numChannels; i++) {
			cvReleaseImage(&(scaledImages[i]));
			scaledImages[i] = cvCreateImage(cvSize((int)(images[i]->width / scale),
				(int)(images[i]->height / scale)), images[i]->depth, images[i]->nChannels);
			cvResize(images[i], scaledImages[i]);
			// Re-compute integral images.
			if ( use_integral_images ) {
				svlCodeProfiler::tic(integ_clock);
				cvReleaseImage(&(intImages[i]));
				cvReleaseImage(&(int2Images[i]));
				intImages[i] = cvCreateImage(cvSize(scaledImages[i]->width + 1, scaledImages[i]->height + 1), IPL_DEPTH_32S, scaledImages[i]->nChannels);
				int2Images[i] = cvCreateImage(cvSize(scaledImages[i]->width + 1, scaledImages[i]->height + 1), IPL_DEPTH_64F, scaledImages[i]->nChannels);
				cvIntegral(scaledImages[i], intImages[i], int2Images[i]);
				svlCodeProfiler::toc(integ_clock);
			}
		}
		cvReleaseImage(&responseImage);
		responseImage = cvCreateImage(cvSize(scaledImages[0]->width, scaledImages[0]->height), IPL_DEPTH_32F, 1);

		svlCodeProfiler::toc(prof3);

		// Erase cached image pointer on GPU to avoid ghosting to this result.
		cm->pitchClearSource(_IMG);
	}

	// free memory
	for ( int i=0; i<numChannels; ++i ) {
		cvReleaseImage(&(scaledImages[i]));
	}

	svlCodeProfiler::toc(prof00);

	return objects;
}
#endif

// Pad all the patches in the dictionary for device convolution, then send to GPU.
void svlCudaSlidingWindowDetector::padAndCachePatches()
{
	if ( cachedLine != NULL && cachedLine->data != NULL )
		cm->freeLine(cachedLine);

	// Create cached pointers for host-side use.
	cachedPatches.reserve(dict._intEntries.size());

	vector<int> paddedSizes;
	paddedSizes.reserve(dict._intEntries.size());
	int total_sz = 0;
	pair<int,int> pad_wh;
	for ( vector<svlIntensityPatch*>::const_iterator p = dict._intEntries.begin(); p != dict._intEntries.end(); ++p ) {
		pad_wh = svlCudaConvolution::paddedPatchSize( (*p)->patch );

		cachedPatches.push_back(CachedPatch());
		cachedPatches.back().w = pad_wh.first;
		cachedPatches.back().h = pad_wh.second;
		cachedPatches.back().N = (*p)->patch->width * (*p)->patch->height; // Original numel.

		// Calculate patch statistics.
		float patchSum = 0.0;
		float patchNorm = 0.0;
		unsigned char *pp = (unsigned char *)(*p)->patch->imageData;
		for ( int y=0; y<(*p)->patch->height; ++y ) {
			for ( int x=0; x<(*p)->patch->width; ++x ) {
				patchSum += (float)pp[x];
				patchNorm += (float)(pp[x] * pp[x]);
			}
			pp += (*p)->patch->widthStep;
		}
		float N = (float)((*p)->patch->width * (*p)->patch->height);
		float patchMean = patchSum / N;
		patchNorm -= patchSum * patchMean;
		cachedPatches.back().mean = patchMean;
		cachedPatches.back().std = sqrt(patchNorm/N);

		paddedSizes.push_back( pad_wh.first * pad_wh.second );
		total_sz += paddedSizes.back();
	}
	total_sz *= 2; // For patches and masks.
	//printf("total size: %d\n", total_sz);

	// Allocate memory.
	float *patchData = (float*)malloc(sizeof(float)*total_sz);
	float *ptr = patchData;
	// Pad patches.
	vector<int>::const_iterator szs = paddedSizes.begin();
	for ( vector<svlIntensityPatch*>::const_iterator p = dict._intEntries.begin(); p != dict._intEntries.end(); ++p ) {
		svlCudaConvolution::padPatchWithMask( (*p)->patch, ptr );
		ptr += *szs * 2; // Add two times patch size: for patch and mask.
		++szs;
	}

	// Allocate device meory.
	cachedLine = cm->allocLine(total_sz);
	// Copy over to GPU.
	cm->lineToGPU(cachedLine, patchData, total_sz);

	// Free host-side memory.
	free(patchData);
	//cm->line2file("line.out", cachedLine);
	//exit(-1);

	// Create individual svlCudaLine pointers.
	float *dev_ptr = cachedLine->data;
	for ( int i=0; i<(int)cachedPatches.size(); ++i ) {
		// Patch line.
		cachedPatches[i].patch = new svlCudaLine();
		cachedPatches[i].patch->w = paddedSizes[i];
		cachedPatches[i].patch->data = dev_ptr;
		dev_ptr += paddedSizes[i];
		// Mask line.
		cachedPatches[i].mask = new svlCudaLine();
		cachedPatches[i].mask->w = paddedSizes[i];
		cachedPatches[i].mask->data = dev_ptr;
		dev_ptr += paddedSizes[i];
		//printf("> %x %x\n", cachedPatches[i].patch->data, cachedPatches[i].mask->data);
	}
}

// Load a decision tree.
void svlCudaSlidingWindowDetector::loadCudaDTree(const char *filename)
{
	if ( dtree != NULL ) {
		delete dtree;
	}
	dtree = new svlCudaDTree(filename);

}


