/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlKdTreeT.h
** AUTHOR(S):   Paul Baumstarck <pbaumstarck@stanford.edu>
**
** DESCRIPTION:
**   Header file for the kd-tree class.
*****************************************************************************/

#pragma once

#if defined(_WIN32)||defined(WIN32)||defined(__WIN32__)
#define _CRT_SECURE_NO_DEPRECATE
#include <limits>
#undef max
#undef min
#endif

#include "svlPoint2d.h"
#include "../vision/svlPoint3d.h"

using namespace std;

template <class P>
struct svlBoxT {
public:
	P max_corner, min_corner;
public:
	inline svlBoxT() {
		max_corner.set_max_pt();
		max_corner *= -1;
		min_corner.set_max_pt();
	}
    inline svlBoxT(const P& pt)
	    : max_corner(P(pt)),
	    min_corner(P(pt)) {}
	inline svlBoxT(const svlBoxT<P>& box)
	    : max_corner(box.max_corner),
	    min_corner(box.min_corner) {}
	inline svlBoxT(const svlBoxT<P>& box1, const svlBoxT<P>& box2)
	    : max_corner(box1.max_corner),
	    min_corner(box1.min_corner) {
	    max_corner.max_with(box2.max_corner);
	    min_corner.min_with(box2.min_corner);
	}
	inline svlBoxT(const P& min_in, const P& max_in)
		: min_corner(min_in), max_corner(max_in) {}
	virtual ~svlBoxT() {}; // do nothing

	inline void addToBox(const P &pt) {
		min_corner.min_with(pt);
		max_corner.max_with(pt);
	}

	inline void unionWithBox(const svlBoxT<P> &box) {
		min_corner.min_with(box.min_corner);
		max_corner.max_with(box.max_corner);
	}

	inline double minDistFromPt(const P &pt) {
		return std::min( (max_corner-pt).pt_min(), (pt-min_corner).pt_min() );
	}

	inline double mDFPWOB(const P &pt, const svlBoxT<P> &box) {
		return minDistFromPtWithOuterBox(pt,box);
	}

	inline double minDistFromPtWithOuterBox(const P &pt, const svlBoxT<P> &box);

	friend inline svlBoxT<P> operator/(const svlBoxT<P>& p, double d) {
		return svlBoxT<P>(p.min_corner, p.min_corner + (p.max_corner-p.min_corner)/2.0);
	}
	
	friend std::ostream& operator<<(std::ostream& os, const svlBoxT& p) {
	    os << "(" << p.min_corner << ") -> (" << p.max_corner << ")";
	    return os;
	}
};

template <class P>
inline double svlBoxT<P>::minDistFromPtWithOuterBox(const P &pt, const svlBoxT<P> &box) {
	// technically just a templated 3D version
	double gd = 1e9;
	for ( int i=0; i<pt.size(); ++i ) {
		if ( max_corner[i] < box.max_corner[i] ) gd = gd < (double)(max_corner[i] - pt[i]) ? gd : (double)(max_corner[i] - pt[i]);
		if ( min_corner[i] > box.min_corner[i] ) gd = gd < (double)(pt[i] - min_corner[i]) ? gd : (double)(pt[i] - min_corner[i]);
	}
	return gd;
}

template <>
inline double svlBoxT<svlPoint2d>::minDistFromPtWithOuterBox(const svlPoint2d &pt, const svlBoxT<svlPoint2d> &box) {
	return std::min( svlPoint2d( max_corner.x == box.max_corner.x ? numeric_limits<double>::max() : max_corner.x - pt.x,
		max_corner.y == box.max_corner.y ? numeric_limits<double>::max() : max_corner.y - pt.y ).pt_min(),
	    svlPoint2d( min_corner.x == box.min_corner.x ? numeric_limits<double>::max() : pt.x - min_corner.x,
		min_corner.y == box.min_corner.y ? numeric_limits<double>::max() : pt.y - min_corner.y ).pt_min() );
}

template <>
inline double svlBoxT<svlPoint3d>::minDistFromPtWithOuterBox(const svlPoint3d &pt, const svlBoxT<svlPoint3d> &box) {
	return std::min( svlPoint3d( max_corner.x == box.max_corner.x ? numeric_limits<double>::max() : max_corner.x - pt.x,
		max_corner.y == box.max_corner.y ? numeric_limits<double>::max() : max_corner.y - pt.y,
		max_corner.z == box.max_corner.z ? numeric_limits<double>::max() : max_corner.z - pt.z ).pt_min(),
	    svlPoint3d( min_corner.x == box.min_corner.x ? numeric_limits<double>::max() : pt.x - min_corner.x,
		min_corner.y == box.min_corner.y ? numeric_limits<double>::max() : pt.y - min_corner.y,
		min_corner.z == box.min_corner.z ? numeric_limits<double>::max() : pt.z - min_corner.z ).pt_min() );
}

