#include "svlDetectorFeatureExtractor.h"
#include "svlVision.h"
#include "svlImageFrame.h"
#include "svlHaarFeatureExtractor.h"
#include "svlCompositeFeatureExtractor.h"

static const char* knownSubclasses[] = {
  "PatchDictionary",
  "CompositeFeatureExtractor",
  "HaarFeatureExtractor",
  NULL
  };

//Helper method used by the two class members (one for string, one for file)
svlDetectorFeatureExtractor* getInstanceHelper(XMLNode &root);

svlDetectorFeatureExtractor*
  svlDetectorFeatureExtractor::getInstance(const string &filename)
{


  //Check that file exists
  ifstream ifs;
  ifs.open(filename.c_str());

  if (ifs.fail())
    {
      SVL_LOG(SVL_LOG_WARNING, "Error while opening detector feature extractor file "<<filename<<": no such file exists, or file is unreadable");
      ifs.close();
      return NULL;
    }



  ifs.close();


  XMLNode root = XMLNode::parseFile(filename.c_str());





  svlDetectorFeatureExtractor * rval =   getInstanceHelper(root);

  if (!rval)
    SVL_LOG(SVL_LOG_WARNING, "...while parsing detector feature extractor file "<<filename)


  return rval;
}

svlDetectorFeatureExtractor*
svlDetectorFeatureExtractor::getInstanceFromString(const char *XML)
{
  XMLNode root = XMLNode::parseString(XML);
  
  svlDetectorFeatureExtractor * rval = getInstanceHelper(root);

  if (!rval)
      SVL_LOG(SVL_LOG_WARNING, "...while getting detector feature extractor instance from string");

  return rval;
}

svlDetectorFeatureExtractor* getInstanceHelper(XMLNode &root)
{
  if (root.isEmpty())
    {
      SVL_LOG(SVL_LOG_WARNING, "Detector feature extractor file could not be parsed");
      return NULL;
    }

  if (!root.getName())
    {
      if (root.nChildNode() != 1)
	{
	  SVL_LOG(SVL_LOG_WARNING, "Detector feature extractor file contains multiple roots");
	  return NULL;
	}

      root = root.getChildNode(0);

      if (!root.getName())
	{
	  SVL_LOG(SVL_LOG_WARNING, "Detector feature extractor file has no named XML root in any of the usual places");
	  return NULL;
	}

    }


  //Try all possible root nodes until we find one that works
  int subclassIdx = 0;
  for(; knownSubclasses[subclassIdx]!=NULL; subclassIdx++)
    {

      //cout << filename << " " << knownSubclasses[subclassIdx] << endl;
      if (!strcmp(root.getName(), knownSubclasses[subclassIdx]))
	  break;
    }

  if (knownSubclasses[subclassIdx] == NULL)
    {
      SVL_LOG(SVL_LOG_WARNING, "No valid xml root found");
      return NULL;
    }

  //SVL_LOG(SVL_LOG_MESSAGE, "Subclass index "<<subclassIdx<<", rootName "<<root.getName());

  //Punt if the file wasn't XML or didn't have the right root node
  svlDetectorFeatureExtractor * rval = NULL;

#define MAKE_DFE_OF_TYPE( TYPE )  rval = new TYPE();
  
  switch(subclassIdx)
    {
    case 0:
       
      MAKE_DFE_OF_TYPE(svlPatchDictionary);
      
      break;
      
    case 1:
      
      MAKE_DFE_OF_TYPE(svlCompositeFeatureExtractor);
      
      break;
           
    case 2:
     
      MAKE_DFE_OF_TYPE(svlHaarFeatureExtractor);

      break;
      
    default:
      SVL_LOG(SVL_LOG_FATAL, "svlDetectorFeatureExtractor passed unrecognized subclass index "<<subclassIdx);
      return NULL;
    }
  
  if (!rval)
    {
      SVL_LOG(SVL_LOG_WARNING, "svlDetectorFeatureExtractor factory: aborting before attempting to read XML");
      return NULL;
    }

  if (!rval->read(root))
    {
      SVL_LOG(SVL_LOG_WARNING, "svlDetectorFeatureExtractor factory: XML reading failed");
      return NULL;
    }

  //SVL_LOG(SVL_LOG_MESSAGE, "Read detector with width "<<rval->windowWidth());
  
  return rval;
}

bool svlDetectorFeatureExtractor::write(const string &filename)
{
  return this->write(filename.c_str());
}

bool svlDetectorFeatureExtractor::write(const char *filename)
{
  ofstream ofs(filename);
  bool rval = this->writeOut(ofs);
  ofs.close();
  return rval;
}

//Default implementation of extract that takes IplImage* channels and
//boxes them up into svlImageFrames
void svlDetectorFeatureExtractor::extract(const vector<CvPoint> &locations, 
					  const vector<IplImage*> &channels, 
					  vector< vector<double> > &output,
					  bool sparse,
					  int outputOffset) const
{
  vector<svlDataFrame*> boxedChannels(channels.size());
  for(int i=0; i<int(channels.size()); i++)
    {
      svlImageFrame *newFrame = new svlImageFrame();
      newFrame->image = channels[i];
      boxedChannels[i] = newFrame;
    }

  this->extract(locations, boxedChannels, output, sparse, outputOffset);

  for(int i=0; i<int(boxedChannels.size()); i++)
    delete boxedChannels[i];
}

void svlDetectorFeatureExtractor::windowResponse(const vector<IplImage *> & channels, vector<double> & output,
						 int outputOffset, const vector<bool> *usedFetures) const
{
  CvPoint location = cvPoint(0,0);
  vector<CvPoint> locations;
  locations.push_back(location);
  vector<vector<double> > outputs;
  

  extract(locations, channels, outputs, true, outputOffset);

  assert(outputs.size() == 1);

  if (output.size())
    {
      if (output.size() < outputOffset + outputs[0].size())
	output.resize(outputOffset + outputs[0].size());

      //TODO-- come up with something better than this slow copy operation
      for (int i =  0; i < outputs[0].size(); i++)
	output[outputOffset + i] = outputs[0][i];
    }
  else
    output = outputs[0];
}

string svlDetectorFeatureExtractor::summary() const
{
  cout << "Method called" << endl;
  return string("[No summary available]");
}
