/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlKdTree.h
** AUTHOR(S):   Paul Baumstarck <pbaumstarck@stanford.edu>
**
** DESCRIPTION:
**   Header file for the kd-tree class.
*****************************************************************************/

#pragma once

#include <iostream>
#include <vector>
#include <queue>
#include <string>
#include <ctime>
#include <limits>
#include "svlBox3d.h"
#include "svlTreeNode.h"
#include "../vision/svlPoint3d.h"

using namespace std;

#define MY_DEBUG 0
#define MY_DEBUG_vx2 0
#define KDT_DEBUG_OUT 0

typedef priority_queue<pair<double,int>, vector<pair<double,int> > > OrdPQ;

class svlKdTree {
protected:
	svlTreeNode *_troot; // the root node of the eponymous tree
	vector<svlPoint3d> *_P; // a pointer to the point cloud to be used in the k-d tree
	vector<svlTreeNode*> _pts2nodes; // lookup the nodes for each point quickly
	vector<double> _max_dists; // maximum distances; accumulated dynamically
	long checks, n_checks, n_levels;
	bool debug, debug1;
	int _bucket_size, n_pts;

public:
	svlKdTree();
	svlKdTree(vector<svlPoint3d> *P_in);
	svlKdTree(vector<svlPoint3d> *P_in, int _bucket_size);
	svlKdTree(vector<svlPoint3d> *P_in, double min_box, vector<svlPoint3d> *P_out);
	svlKdTree(vector<svlPoint3d> *P_in, double min_box, vector<svlPoint3d> *P_out, vector<int> *in2out = NULL);
	svlKdTree(vector<svlPoint3d> *P_in, double min_box, vector<svlPoint3d> *P_out, int bucket_size, vector<int> *in2out = NULL);
	virtual ~svlKdTree();
	void createKdTree(vector<svlPoint3d> *P_in);
	void createKdTree(vector<svlPoint3d> *P_in, int _bucket_size);
	void createKdTreeDec(vector<svlPoint3d> *P_in, double min_box, vector<svlPoint3d> *P_out, vector<int> *in2out = NULL);
	void createKdTreeDec(vector<svlPoint3d> *P_in, double min_box, vector<svlPoint3d> *P_out, int bucket_size, vector<int> *in2out = NULL);
	void rebucket(int bucket_size);

	// printing and checking functions
	int printNodeAtDepth(svlTreeNode *parent, svlTreeNode *node, int ct, int dep, string str);
	void printTree();
	bool verifyKdTree();

	// searching functinos
	void findKNN(int ix, int k, double max_dist, vector<int> &ixs);
	void findKNN(int ix, int k, double max_dist, vector<pair<double,int> > &dist_ixs);
	void findAllKNN(int k, double max_dist, vector<vector<int> > &knn);
	void findKNN(svlPoint3d &pt, int k, double max_dist, vector<int> &ixs);
	void findKNN(svlPoint3d &pt, int k, double max_dist, vector<pair<double,int> > &dist_ixs);

	// bureaucracy functions
	void start_debug();
	void stop_debug();
	void print_debug();
	void set_bucket_size(int bucket_size);
	int get_bucket_size();
	void testFindExactAll();
	int get_n_pts();

	bool searchDuplicateIxs(); // verifies there are no duplicated indices in the tree
protected:
	// searching functions
	int findExactNode(int &ix, svlPoint3d &pt, svlTreeNode* &node, int dep);
	void findKNNd(int ix, int k, double max_dist, OrdPQ &pq);
	void findKNN(int ix, int k, double max_dist, OrdPQ &pq);
	void findKNN(svlPoint3d &pt, int k, double max_dist, OrdPQ &pq);
	void findKNNbucket1(int ix, int k, double max_dist, OrdPQ &pq);
	
	// tree creation functions
	int select_median(vector<int> &ix, int left, int right, int med, int dim);
	int select_best_median(vector<int> &ix, int left, int right, int &dim, double &span);
	int partition(vector<int> &ix, int left, int right, int piv, int dim);
	svlTreeNode* createKdSubTree(vector<int> &ix, int left, int right);
	svlTreeNode* createKdSubTreeDec(vector<int> &ix, int left, int right, double min_box, vector<svlPoint3d> *P_out, vector<int> *in2out = NULL);
	
	// util functions
	void deleteNodes(svlTreeNode* node);
	void connectParents(svlTreeNode *parent, svlTreeNode *node);
	bool verifyKdTree(svlTreeNode *node);
	int getDepthOfNode(svlTreeNode* &q);
	void fillBoundingBoxes(svlTreeNode *q);
	bool isLeaf(svlTreeNode *q);
	svlTreeNode* getSibling(svlTreeNode *q);

	inline void point2PQ(svlPoint3d &pt, int &k, OrdPQ &pq, double &max_dist, int index);
};

