/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlSortVectorDI.cpp
** AUTHOR(S):   Paul Baumstarck <pbaumstarck@stanford.edu>
** DESCRIPTION:
**   Indexed vector for sorting and recovering the permutation. Similar to
**   MATLAB [vals_sorted,idx] = sort(vals);
*****************************************************************************/

#include "svlSortVectorDI.h"

svlSortVectorDI::svlSortVectorDI() : _ix(0), _sorted(false), _ascending(true) {
}

svlSortVectorDI::svlSortVectorDI(bool _ascending) : _ix(0), _sorted(false), _ascending(_ascending) {
}

svlSortVectorDI::svlSortVectorDI(int reserved_size) : _ix(0), _sorted(false), _ascending(true) {
	_myvec.reserve(reserved_size);
}

svlSortVectorDI::svlSortVectorDI(int reserved_size, bool _ascending) : _ix(0), _sorted(false), _ascending(_ascending) {
	_myvec.reserve(reserved_size);
}

svlSortVectorDI::~svlSortVectorDI() {
	_myvec.clear();
}

bool svlSortVectorDI::is_sorted() {
	return _sorted;
}

void svlSortVectorDI::reserve(int reserved_size) {
	_myvec.reserve(reserved_size);
}

double svlSortVectorDI::get_value(int ix) {
	return _myvec[ix].first;
}

int svlSortVectorDI::get_ix(int ix) {
	return _myvec[ix].second;
}

int svlSortVectorDI::get_orig_ix(int ix) {
	return _orig[ix];
}

double svlSortVectorDI::get_orig_val(int ix) {
	if ( _sorted ) {
		assert(_orig[ix]!=-1);
		return _myvec[_orig[ix]].first;
	} else {
		return _myvec[ix].first;
	}
}

void svlSortVectorDI::setAscending(bool ascending) {
	_ascending = ascending;
}

bool svlSortVectorDI::getAscending() {
	return _ascending;
}


void svlSortVectorDI::setDescending(bool descending) {
	_ascending = !descending;
}


bool svlSortVectorDI::getDescending() {
	return !_ascending;
}

int svlSortVectorDI::size() {
	return (int)_myvec.size();
}


void svlSortVectorDI::push_back(double val) {
	_myvec.push_back(make_pair(val,_ix));
	++_ix;
	_sorted = false;
}

void svlSortVectorDI::pop_back() {
	_orig[_myvec.back().second] = -1; // invalidate original indexing
	_myvec.pop_back();
}

void svlSortVectorDI::clear() {
	_myvec.clear();
	_ix = 0;
	_sorted = false;
}

void svlSortVectorDI::sort() {
	if ( _ascending )
		std::sort(_myvec.begin(), _myvec.end());
	else {
		greater<pair<double,int> > comp;
		std::sort(_myvec.begin(), _myvec.end(), comp);
	}
	_ix = (int)_myvec.size();
	_orig = vector<int>(_ix,-1);
	for ( int i=0; i<(int)_myvec.size(); ++i ) { // recovery indices
		if ( _myvec[i].second != -1 ) {
 			_orig[_myvec[i].second] = i;
		}
	}
	_sorted = true;
}

void svlSortVectorDI::reorder() { // reorder the indices (ix's)
	_ix = 0;
	for ( vector<pair<double,int> >::iterator i=_myvec.begin(); i!=_myvec.end(); ++i ) {
		(i->second) = _ix;
		++_ix;
	}
}


double svlSortVectorDI::at(int _ix) {
	return _myvec[_ix].first;
}


int svlSortVectorDI::ix_at(int _ix) {
	return _myvec[_ix].second;
}


double svlSortVectorDI::val_at(int _ix) {
	return _myvec[_ix].first;
}


pair<double,int>& svlSortVectorDI::operator[](int ix) {
	return _myvec[ix];
}


pair<double,int>& svlSortVectorDI::back() {
	return _myvec.back();
}


void svlSortVectorDI::erase(int i) {
	if ( i >= (int)_myvec.size() ) return;
	vector<pair<double,int> >::iterator it = _myvec.begin();
	for ( int j=0; i<j; ++j, ++it );
	_myvec.erase(it);
	_orig[it->second] = -1; // invalidate original indexing
	// does not invalidate the state of _sorted
}

