/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlEdgeFeatures.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**
*****************************************************************************/

#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <vector>
#include <cassert>

#include "cv.h"
#include "cxcore.h"

#include "svlBase.h"
#include "svlEdgeFeatures.h"
#include "svlVisionUtils.h"

using namespace std;

svlEdgeFeatures::svlEdgeFeatures(IplImage *img) :
    _image(NULL), _gX(NULL), _gY(NULL)
{
    // set default options
    // TODO: add some options

    if (img != NULL) setImage(img);
}

svlEdgeFeatures::~svlEdgeFeatures()
{
    if (_image != NULL) {
        cvReleaseImage(&_image);
        cvReleaseImage(&_gX);
        cvReleaseImage(&_gY);
    }
}

// set image for edge features
void svlEdgeFeatures::setImage(IplImage *img)
{
    if (_image != NULL) {
        cvReleaseImage(&_image); _image = NULL;
        cvReleaseImage(&_gX);    _gX = NULL;
        cvReleaseImage(&_gY);    _gY = NULL;
    }

    if (img == NULL) return;

    // cache image and compute gradient information
    _image = greyImage(img);
    computeGradients();
}

// compute feature responses
vector<double> svlEdgeFeatures::computeFeature(const vector<CvPoint>& e)
{
    SVL_ASSERT((_image != NULL) && (!e.empty()));
    vector<double> fv;

    // create vector of x's and y's
    vector<double> x(e.size());
    vector<double> y(e.size());
    vector<double> gx(e.size());
    vector<double> gy(e.size());
    for (unsigned i = 0; i < e.size(); i++) {
        x[i] = (double)e[i].x / (double)_image->width;
        y[i] = (double)e[i].y / (double)_image->height;
        gx[i] = (double)(CV_IMAGE_ELEM(_gX, float, e[i].y, e[i].x) *
            CV_IMAGE_ELEM(_gX, float, e[i].y, e[i].x));
        gy[i] = (double)(CV_IMAGE_ELEM(_gY, float, e[i].y, e[i].x) *
            CV_IMAGE_ELEM(_gX, float, e[i].y, e[i].x));
    }

    pair<double, double> r;

    // edge extent
    r = range(x);
    fv.push_back(r.first);
    fv.push_back(r.second);
    r = range(y);
    fv.push_back(r.first);
    fv.push_back(r.second);

    // angle and line fit
    double a, b, mse;
    if (e.size() > 1) {
        mse = svlLineFit(e, a, b);
    } else {
        // some edges only have one pixel
        a = 0.0; b = 0.0; mse = 0.0;
    }
    fv.push_back(a / (double)_image->width);
    fv.push_back(b / (double)_image->height);
    fv.push_back(atan2(b, a));
    fv.push_back(mse / (_image->width * _image->height));

    // lengths
    fv.push_back((double)e.size()); // num points
    fv.push_back((fv[1] - fv[0]) * (fv[1] - fv[0]) +
        (fv[3] - fv[2]) * (fv[3] - fv[2])); // sqr length?
    fv.push_back((fv[1] - fv[0]) * (fv[3] - fv[2])); // area
    
    // gradient range, mean and variance along edgel
    r = range(gx);
    fv.push_back(r.first);
    fv.push_back(r.second);
    r = range(gy);
    fv.push_back(r.first);
    fv.push_back(r.second);
    fv.push_back(mean(gx));
    fv.push_back(mean(gy));
    fv.push_back(variance(gx));
    fv.push_back(variance(gy));

    return fv;
}

vector<vector<double> > svlEdgeFeatures::computeFeatures(const vector<vector<CvPoint> >& e)
{
    SVL_ASSERT(_image != NULL);
    vector<vector<double> > fv(e.size());

    for (unsigned i = 0; i < e.size(); i++) {
        fv[i] = computeFeature(e[i]);
    }

    return fv;
}

// compute graidents
void svlEdgeFeatures::computeGradients()
{
    SVL_ASSERT(_image != NULL);
    if (_gX != NULL) {
        cvReleaseImage(&_gX);
        cvReleaseImage(&_gY);
    }

    _gX = cvCreateImage(cvGetSize(_image), IPL_DEPTH_16S, 1);
    _gY = cvCreateImage(cvGetSize(_image), IPL_DEPTH_16S, 1);
    cvSobel(_image, _gX, 1, 0, 3);
    cvSobel(_image, _gY, 0, 1, 3);
    
    convertInPlace(&_gX, IPL_DEPTH_32F, 1.0 / 256.0);
    convertInPlace(&_gY, IPL_DEPTH_32F, 1.0 / 256.0);
    cvSmooth(_gY, _gX, CV_BLUR, 5, 5);
    cvSmooth(_gY, _gX, CV_BLUR, 5, 5);
}


