/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlPatchDefinition.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**              Sid Batra<sidbatra@cs.stanford.edu>
**              Olga Russakovsky <olga@cs.stanford.edu>
**              Ian Goodfellow <ia3n@cs.stanford.edu>
**
*****************************************************************************/

#include <cassert>
#include <iostream>
#include <fstream>
#include <vector>
#include <limits>

#include "cv.h"
#include "cxcore.h"
#include "highgui.h"

#include "svlBase.h"
#include "svlVision.h"

#include "svlPatchDefinition.h"
#include "svlAbsPatchMaskCreator.h"
#include "svlDevel.h"

using namespace std;

map<IplImage*,svlPatchStats*> svlIntensityPatchDefinition::patch_cache = map<IplImage*,svlPatchStats*>();

// svlPatchDefinition class --------------------------------------------------

svlPatchDefinition::svlPatchDefinition(svlPatchDictionary * parent) :
  _template(NULL), _validChannel(0), _parent(parent)
{
    _validRect = cvRect(0, 0, 0, 0);
}

svlPatchDefinition::svlPatchDefinition(XMLNode& node, svlPatchDictionary * parent) :
  _template(NULL), _validChannel(0), _parent(parent)
{
    _validRect = cvRect(0, 0, 0, 0);
    //read(node);
}

svlPatchDefinition::svlPatchDefinition(const IplImage *t, const CvRect& rect, \
				       svlPatchDictionary * parent, int channel) \
  : _parent(parent)
{
  if (t != NULL) {
        _template = cvCloneImage(t);
    } else {
        _template = NULL;
    }
    _validRect.x = rect.x;
    _validRect.y = rect.y;
    _validRect.width = rect.width;
    _validRect.height = rect.height;
    _validChannel = channel;
}

svlPatchDefinition::~svlPatchDefinition()
{
    if (_template != NULL) {
        cvReleaseImage(&_template);
	_template = NULL;
    }
}

// constuction helper functions
svlSmartPointer<svlPatchDefinition> svlPatchDefinition::createPatchDefinition(svlPatchDefinitionType t, svlPatchDictionary * parent)
{
    svlSmartPointer<svlPatchDefinition> p;
    switch (t) {
    case SVL_INTENSITY_PATCH:
        p = svlSmartPointer<svlPatchDefinition>(new svlIntensityPatchDefinition(parent));
        break;
    case SVL_DEPTH_PATCH:
        p = svlSmartPointer<svlPatchDefinition>(new svlDepthPatchDefinition(parent));
        break;
    case SVL_ABS_PATCH:
        p = svlSmartPointer<svlPatchDefinition>(new svlAbsPatchDefinition(parent));
        break;
    }

    return p;
}

svlSmartPointer<svlPatchDefinition> svlPatchDefinition::createPatchDefinition(XMLNode &node, svlPatchDictionary * parent)
{
    assert(!node.isEmpty());
    svlSmartPointer<svlPatchDefinition> p;

    const char *t = node.getAttribute("type");
    assert(t != NULL);

    if (!strcasecmp(t, "intensity")) {
      p = svlSmartPointer<svlPatchDefinition>(new svlIntensityPatchDefinition(node, parent));
    } else if (!strcasecmp(t, "depth")) { 
      p = svlSmartPointer<svlPatchDefinition>(new svlDepthPatchDefinition(node,parent));
    } else if (!strcasecmp(t, "abs")) {
      p = svlSmartPointer<svlPatchDefinition>(new svlAbsPatchDefinition(node,parent));
    } else {
        SVL_LOG(SVL_LOG_FATAL, "unknown patch type " << t);
    }

    return p;
}


bool svlPatchDefinition::read(XMLNode &node)
{		
	
    assert(!node.isEmpty());
    if (_template != NULL) {
        cvReleaseImage(&_template);
    }

    // read valid region
    if (node.getAttribute("validRect")) {
        vector<int> v;
        parseString<int>(node.getAttribute("validRect"), v);
        assert(v.size() == 4);
        _validRect = cvRect(v[0], v[1], v[2], v[3]);
    } else {
        _validRect = cvRect(0, 0, 0, 0);
    }

    // read channel
    if (node.getAttribute("channel")) {
        _validChannel = atoi(node.getAttribute("channel"));
    } else {
        _validChannel = 0;
    }

    // read patch size
    int w = 0;
    int h = 0;
    if (node.getAttribute("size")) {
        vector<int> v;
        parseString<int>(node.getAttribute("size"), v);
        assert(v.size() == 2);
        w = v[0]; h = v[1];
    }

    // read patch data
    if ((w > 0) && (h > 0)) {
        _template = cvCreateImage(cvSize(w, h), IPL_DEPTH_8U, 1);
        vector<int> v;
        parseString(node.getText(), v);
        assert(v.size() == (unsigned)(w * h));
        unsigned n = 0;
        for (unsigned y = 0; y < (unsigned)_template->height; y++) {
            char *p = &_template->imageData[y * _template->widthStep];
            for (unsigned x = 0; x < (unsigned)_template->width; x++) {
                p[x] = (char)v[n++];
            }
        }
    }
	
    return true;
}


bool svlPatchDefinition::write(ostream& os)
{
    assert(_template != NULL);

    os << "  <PatchDefinition type=\"";
    switch (this->patchType()) 
	{
    case SVL_INTENSITY_PATCH:
        os << "intensity";
        break;
    case SVL_DEPTH_PATCH:
        os << "depth";
        break;
    case SVL_ABS_PATCH:
        os << "abs";
        break;
    default:
        SVL_LOG(SVL_LOG_FATAL, "unknown patch type");
    }
    os << "\"\n"
       << "    validRect=\"" << _validRect.x << " " << _validRect.y << " "
       << _validRect.width << " " << _validRect.height << "\"\n"
       << "    channel=\"" << _validChannel << "\"\n"
       << "    size=\"" << _template->width << " " << _template->height << "\">\n";
    
    for (unsigned y = 0; y < (unsigned)_template->height; y++) 
	{
        os << "   ";
        for (unsigned x = 0; x < (unsigned)_template->width; x++) 
		{
            if (_template->depth == IPL_DEPTH_8U) 
			{
                os << " " << (int)CV_IMAGE_ELEM(_template, unsigned char, y, x);
            } 
			else if (_template->depth == IPL_DEPTH_32F) 
			{
                os << " " << CV_IMAGE_ELEM(_template, float, y, x);
            }
			else
			{
                SVL_LOG(SVL_LOG_FATAL, "unrecognized template type");
            }
        }        
        os << "\n";
    }
    
    os << "  </PatchDefinition>\n";

    return true;
}

static double correlationResponse(IplImage *img, IplImage *tmpl) {
  IplImage *response = cvCreateImage(cvSize(img->width - tmpl->width + 1,
					    img->height - tmpl->height + 1),
				     IPL_DEPTH_32F, 1);
  // do not subtract the mean of each patch, just do
  // normalized cross-correlation (so expansion doesn't
  // affect it negatively)
  cvMatchTemplate(img, tmpl, response, CV_TM_CCORR_NORMED);  

  double min_value, max_value;
  cvMinMaxLoc(response, &min_value, &max_value);

  cvReleaseImage(&response);
  return max_value;
}

static double correlationResponseWithExpansion(IplImage *imBiggerWidth, IplImage *imBiggerHeight) {

  int w_small = imBiggerHeight->width;
  int w_large = imBiggerWidth->width;

  int h_small = imBiggerWidth->height;
  int h_large = imBiggerHeight->height;

  // first, expand in height
  IplImage *expanded = cvCreateImage(cvSize(w_large, h_large * 2 - h_small),
				     imBiggerWidth->depth, imBiggerWidth->nChannels);

  cvZero(expanded);
  cvSetImageROI(expanded, cvRect(0, h_large - h_small, w_large, h_small));
  cvCopy(imBiggerWidth, expanded);
  cvResetImageROI(expanded);

  double val1 = correlationResponse(expanded, imBiggerHeight);
  cvReleaseImage(&expanded);

  // now expand in width
  expanded = cvCreateImage(cvSize(w_large * 2 - w_small, h_large),
			   imBiggerHeight->depth, imBiggerHeight->nChannels);
  cvZero(expanded);
  cvSetImageROI(expanded, cvRect(w_large - w_small, 0, w_small, h_large));
  cvCopy(imBiggerHeight, expanded);
  cvResetImageROI(expanded);

  double val2 = correlationResponse(expanded, imBiggerWidth);
  cvReleaseImage(&expanded);

  //cerr << " (" << val1 << ", " << val2 << ")";

  return max(val1, val2);
}

static bool enoughOverlap(CvRect rect1, int w1, int h1, CvRect rect2, int w2, int h2) {
  // coordinates of the overlap area
  int x1 = max(rect1.x, rect2.x);
  int y1 = max(rect1.y, rect2.y);
  int x2 = min(rect1.x + rect1.width + w1, rect2.x + rect2.width + w2);
  int y2 = min(rect1.y + rect1.height + h1, rect2.y + rect2.height + h2);

  return ((x2 - x1) >= min(w1, w2)) && ((y2-y1) >= min(h1, h2));
}

bool svlPatchDefinition::isCorrelatedWith(svlPatchDefinition *patch, double threshold) const {
  // checks if they correspond to the same region within the image; if they don't,
  // then they're definitely not correlated

  // assumes that both are valid entries, and thus are non-uniform
  // also assumes that they're of the same type

  if (!isSameType(patch)) {
    //cerr << "not same type" << endl;
    return false;
  }

  if (_template->nChannels != patch->_template->nChannels) {
    return false;
  }

  int a_width = _template->width;
  int b_width = patch->_template->width;

  int a_height = _template->height;
  int b_height = patch->_template->height;

  //cerr << a_width << "x" << a_height << ", " << b_width << "x" << b_height << ": ";

  // if the templates come from completely different areas of the image,
  // they're obviously not correlated
  // ("completely different" is defined as even if the two patches
  // are positioned with as much overlap as possible within the valid
  // rectangle, they still overlap by less than min(w1, w2) x min(h1, h2))
  if (!enoughOverlap(_validRect, a_width, a_height,
		     patch->_validRect, b_width, b_height)) {
    //cerr << "not enough overlap" << endl;
    return false;
  }


  double value = 0;

  if (a_width <= b_width && a_height <= b_height) {
    // patch->_template strictly bigger
    //cerr << "patch strictly bigger, ";
    value = correlationResponse(patch->_template, _template);
  
  } else if (b_width <= a_width && b_height <= a_height) {
    // template strictly bigger
    //cerr << "templ strictly bigger, ";
    value = correlationResponse(_template, patch->_template);
  
  } else if (a_width < b_width) {
    // patch is bigger in width, but template is bigger in height
    //cerr << "patch bigger in w, ";
    value = correlationResponseWithExpansion(patch->_template, _template);
  
  } else if (b_width < a_width) {
    // template is strictly bigger in width
    //cerr << "patch bigger in h, ";
    value = correlationResponseWithExpansion(_template, patch->_template);
  
  } else {
    assert(false); // debugging to make sure all cases are considered
  }

  //cerr << value << endl;
  return (value > threshold);
}

IplImage *svlPatchDefinition::responseImage(const std::vector<IplImage *>& images) const
{
    assert(images.size() > (unsigned)_validChannel);
    return responseImage(images[_validChannel]);
}

/*double svlPatchDefinition::patchValue(const IplImage *responseImage, const CvPoint& location)
{
  SVL_ASSERT_MSG(responseImage->depth == IPL_DEPTH_32F,"in svlPatchDefinition::patchValue fn");
  SVL_ASSERT_MSG(responseImage->nChannels == 1, "in svlPatchDefinition::patchValue fn");

#if 0
    assert((location.x + _validRect.x + _validRect.width <= responseImage->width) &&
        (location.y + _validRect.y + _validRect.height <= responseImage->height));
#endif

    float maxScore = -numeric_limits<float>::max();

    float *p = &CV_IMAGE_ELEM(responseImage, float, location.y + _validRect.y,
        location.x + _validRect.x);

    size_t inc = responseImage->widthStep / sizeof(float);
    for (int dy = 0; dy < _validRect.height; dy++) {
        for (int dx = 0; dx < _validRect.width; dx++) {
            if (p[dx] > maxScore) {
                maxScore = p[dx];
            }
        }        
        p += inc;
    }

    return (double)maxScore;
    }*/

double svlPatchDefinition::patchValue(const IplImage *responseImage, const CvPoint& location) const
{
  SVL_ASSERT_MSG(responseImage->depth == IPL_DEPTH_32F,"in svlPatchDefinition::patchValue fn");
  SVL_ASSERT_MSG(responseImage->nChannels == 1, "in svlPatchDefinition::patchValue fn");

#if 1
  if(! ((location.x + _validRect.x + _validRect.width <= responseImage->width) &&
	(location.y + _validRect.y + _validRect.height <= responseImage->height))) {
    cerr << "location.x: " << location.x << endl;
    cerr << "_validRect.x: " << _validRect.x << endl;
    cerr << "_validRect.width: " << _validRect.width << endl;
    cerr << "responseImage->width: " << responseImage->width << endl;
        
    cerr << "location.y: " << location.y << endl;
    cerr << "_validRect.y: " << _validRect.y << endl;
    cerr << "_validRect.height: " << _validRect.height << endl;
    cerr << "responseImage->height: " << responseImage->height << endl;
    cerr << endl;
  }
#endif

    cvSetImageROI((IplImage *)responseImage, cvRect(location.x + _validRect.x,
						    location.y + _validRect.y, 
						    _validRect.width,
						    _validRect.height));
    double min_value, max_value;
    cvMinMaxLoc(responseImage, &min_value, &max_value);
    cvResetImageROI((IplImage *)responseImage);

    return (double) max_value;
}

vector<double> svlPatchDefinition::patchValues(const IplImage *image,
    const vector<CvPoint>& locations)
{
    assert(false);
    return vector<double>();
}

vector<double> svlPatchDefinition::patchValues(const vector<IplImage *>& images,
    const vector<CvPoint>& locations)
{
    assert(images.size() > (unsigned)_validChannel);
    return patchValues(images[_validChannel], locations);
}

double svlPatchDefinition::trainingResponseValue(IplImage * image) const
{
  IplImage * response = responseImage(image);
  double rval = patchValue(response, cvPoint(0,0));
  cvReleaseImage(&response);

  return rval;
}


// svlIntensityPatchDefinition class --------------------------------------------------

svlIntensityPatchDefinition::svlIntensityPatchDefinition( svlPatchDictionary * parent) :
    svlPatchDefinition(parent)
{
    // do nothing
}

svlIntensityPatchDefinition::svlIntensityPatchDefinition(XMLNode& node, svlPatchDictionary * parent) :
  svlPatchDefinition(node, parent)
{
    // do nothing
	read(node);
}

svlIntensityPatchDefinition::svlIntensityPatchDefinition(const IplImage *t, const CvRect& rect, svlPatchDictionary * parent, int channel) :
  svlPatchDefinition(t, rect, parent, channel)
{
    // do nothing

}

svlIntensityPatchDefinition::~svlIntensityPatchDefinition()
{
    // do nothing
}


bool svlIntensityPatchDefinition::read(XMLNode &node)
{		
	
    assert(!node.isEmpty());
    if (_template != NULL) {
        cvReleaseImage(&_template);
    }

    // read valid region
    if (node.getAttribute("validRect")) {
        vector<int> v;
        parseString<int>(node.getAttribute("validRect"), v);
        assert(v.size() == 4);
        _validRect = cvRect(v[0], v[1], v[2], v[3]);
    } else {
        _validRect = cvRect(0, 0, 0, 0);
    }

    // read channel
    if (node.getAttribute("channel")) {
        _validChannel = atoi(node.getAttribute("channel"));
    } else {
        _validChannel = 0;
    }

    // read patch size
    int w = 0;
    int h = 0;
    if (node.getAttribute("size")) {
        vector<int> v;
        parseString<int>(node.getAttribute("size"), v);
        assert(v.size() == 2);
        w = v[0]; h = v[1];
    }

    // read patch data
    if ((w > 0) && (h > 0)) {
        _template = cvCreateImage(cvSize(w, h), IPL_DEPTH_8U, 1);
        vector<int> v;
        parseString(node.getText(), v);
        assert(v.size() == (unsigned)(w * h));
        unsigned n = 0;
        for (unsigned y = 0; y < (unsigned)_template->height; y++) {
            char *p = &_template->imageData[y * _template->widthStep];
            for (unsigned x = 0; x < (unsigned)_template->width; x++) {
                p[x] = (char)v[n++];
            }
        }
    }
	
    return true;
}

// Read a patch and return as a struct pointer.
svlIntensityPatch* svlIntensityPatchDefinition::readIntensityPatch(XMLNode &node)
{
	svlIntensityPatch *ret = new svlIntensityPatch();

    // Read valid region
    if ( node.getAttribute("validRect") ) {
        vector<int> v;
        parseString<int>(node.getAttribute("validRect"), v);
        assert(v.size() == 4);
		ret->rect = cvRect(v[0], v[1], v[2], v[3]);
    } else {
        ret->rect = cvRect(0, 0, 0, 0);
    }

    // Read channel
    if ( node.getAttribute("channel") ) {
        ret->channel = atoi(node.getAttribute("channel"));
    } else {
        ret->channel = 0;
    }

    // Read patch size
    int w = 0;
    int h = 0;
    if (node.getAttribute("size")) {
        vector<int> v;
        parseString<int>(node.getAttribute("size"), v);
        assert(v.size() == 2);
        w = v[0]; h = v[1];
    }

    // Read patch data
    if ( w > 0 && h > 0 ) {
        ret->patch = cvCreateImage(cvSize(w, h), IPL_DEPTH_8U, 1);
        vector<int> v;
        parseString(node.getText(), v);
        assert(v.size() == (unsigned)(w * h));
        unsigned n = 0;
        for (unsigned y = 0; y < (unsigned)ret->patch->height; y++) {
            char *p = &ret->patch->imageData[y * ret->patch->widthStep];
            for (unsigned x = 0; x < (unsigned)ret->patch->width; x++) {
                p[x] = (char)v[n++];
            }
        }
    }
	
    return ret;
}

IplImage *svlIntensityPatchDefinition::responseImage(const IplImage *image) const
{
  SVL_ASSERT_MSG(image->nChannels == 1,
		 "in svlIntensityPatch::responseImage actually " << image->nChannels << " channels");

    static int handle = svlCodeProfiler::getHandle("svlIntensityPatchDefinition::responseImage (simple)");
    svlCodeProfiler::tic(handle);

    assert((_template != NULL));
    assert((image != NULL));

    int twidth = _template->width;
    int iwidth = image->width;
    int width = iwidth - twidth +1;
    int height = image->height - _template->height+1;


    IplImage *response = cvCreateImage(cvSize(width,
	    height), IPL_DEPTH_32F, 1);
    cvMatchTemplate(image, _template, response, CV_TM_CCOEFF_NORMED);
    
    svlCodeProfiler::toc(handle);
    return response;
}


double svlIntensityPatchDefinition::trainingResponseValue(IplImage * image) const
{


  //responseImage
  SVL_ASSERT_MSG(image->nChannels == 1,
		 "in svlIntensityPatch::trainingResponseValue actually " << image->nChannels << " channels");


    assert((_template != NULL));
    assert((image != NULL));

    int twidth = _template->width;
    int theight = _template->height;
    //int iwidth = image->width;
    //int iheight = image->height;

    cvSetImageROI(image, cvRect(_validRect.x, _validRect.y, _validRect.width+twidth-1, _validRect.height+theight-1));

    //int width = iwidth - twidth +1;
    //int height = image->height - _template->height+1;


    IplImage *responseImage = cvCreateImage(cvSize(_validRect.width, _validRect.height), IPL_DEPTH_32F,1);
//cvCreateImage(cvSize(width, height), IPL_DEPTH_32F, 1);

    //cvShowImage("image",image);

    //cvShowImage("template",_template);

    //cvShowImage("responseImage",responseImage);
    //cvWaitKey(0);

    cvMatchTemplate(image, _template, responseImage, CV_TM_CCOEFF_NORMED);


    cvResetImageROI(image);
    

    //patchValue
    SVL_ASSERT_MSG(responseImage->depth == IPL_DEPTH_32F,"in svlPatchDefinition::patchValue fn");
    SVL_ASSERT_MSG(responseImage->nChannels == 1, "in svlPatchDefinition::patchValue fn");

#if 1  
    if(! (( _validRect.width <= responseImage->width) &&
	  ( _validRect.height <= responseImage->height)))
      {
	cout << "In trainingResponseValue:"<<endl;
	cerr << "location.x: " << 0 << endl;
	cerr << "_validRect.x: " << _validRect.x << endl;
	cerr << "_validRect.width: " << _validRect.width << endl;
	cerr << "responseImage->width: " << responseImage->width << endl;
        
	cerr << "location.y: " << 0 << endl;
	cerr << "_validRect.y: " << _validRect.y << endl;
	cerr << "_validRect.height: " << _validRect.height << endl;
	cerr << "responseImage->height: " << responseImage->height << endl;
	cerr << endl;
      }
#endif


 


    //cout << "Using a "<<_validRect.width<<"x"<<_validRect.height<<" patch of a "<<responseImage->width<<"x"<<responseImage->height<<" image"<<endl;


    
    /*cvSetImageROI((IplImage *)responseImage, cvRect( _validRect.x,
						     _validRect.y, 
						    _validRect.width,
						    _validRect.height));*/
    double min_value, max_value;
    cvMinMaxLoc(responseImage, &min_value, &max_value);
    //cvResetImageROI((IplImage *)responseImage);

    cvReleaseImage(&responseImage);


    return (double)  max_value;
}


IplImage *svlIntensityPatchDefinition::responseImage(const IplImage *image, 
    const IplImage *imageSum, const IplImage *imageSumSq) const
{
  SVL_ASSERT_MSG(image->nChannels == 1,
		 "in svlIntensityPatch::responseImage actually " << image->nChannels << " channels");
  SVL_ASSERT_MSG(imageSum->nChannels == 1,
		 "in svlIntensityPatchDefinition::responseImage fn");
  SVL_ASSERT_MSG(imageSumSq->nChannels == 1,
		 "in svlIntensityPatchDefinition::responseImage fn");

    static int handle = svlCodeProfiler::getHandle("svlIntensityPatchDefinition::responseImage");
    svlCodeProfiler::tic(handle);

    assert((_template != NULL) && (image != NULL) &&
        (imageSum != NULL) && (imageSumSq != NULL));

    IplImage *response = cvCreateImage(cvSize(image->width - _template->width + 1,
	    image->height - _template->height + 1), IPL_DEPTH_32F, 1);	

	float patchSum = 0.0;
    float patchNorm = 0.0;
    unsigned char *pp = (unsigned char *)_template->imageData;
    for (int y = 0; y < _template->height; y++) {
        for (int x = 0; x < _template->width; x++) {
            patchSum += (float)pp[x];
            patchNorm += (float)(pp[x] * pp[x]);
        }
        pp += _template->widthStep;
    }
    float N = (float)(_template->width * _template->height);
    float patchMean = patchSum / N;
    patchNorm -= patchSum * patchMean; 

    /*****/
    IplImage *convolution_response = cvCreateImage(cvSize(image->width - _template->width + 1,
							  image->height - _template->height + 1), IPL_DEPTH_32F, 1);
    cvMatchTemplate(image, _template, convolution_response, CV_TM_CCORR);
    /***/

	// Output terms in summation.
	/*float *num = (float*)malloc(sizeof(float)*convolution_response->height*convolution_response->width);
	float *den = (float*)malloc(sizeof(float)*convolution_response->height*convolution_response->width);
	float *n = num, *d = den;*/

    //int numMultiplications = 0;
    for (int y = 0; y < image->height - _template->height + 1; y++) {
        const unsigned char *p = &CV_IMAGE_ELEM(image, unsigned char, y, 0);
        const int *ps = &CV_IMAGE_ELEM(imageSum, int, y, 0);
        const int *ps2 = &CV_IMAGE_ELEM(imageSum, int, y + _template->height, 0);
        const double *psSq = &CV_IMAGE_ELEM(imageSumSq, double, y, 0);
        const double *psSq2 = &CV_IMAGE_ELEM(imageSumSq, double, y + _template->height, 0);
        float *q = &CV_IMAGE_ELEM(response, float, y, 0);
        for (int x = 0; x < image->width - _template->width + 1; x++, p++) {
            float windowSum = (float)(ps[x] + ps2[x + _template->width] - 
                ps[x + _template->width] - ps2[x]);
            float windowNorm = (float)(psSq[x] + psSq2[x + _template->width] - 
                psSq[x + _template->width] - psSq2[x]);

	    /***/
            /*const unsigned char *pp = (const unsigned char *)_template->imageData;
            const unsigned char *pw = p;

	    unsigned int convolutionSum = 0;
            for (int v = 0; v < _template->height; v++) {
                for (int u = 0; u < _template->width; u++) {
                    convolutionSum += pp[u] * pw[u];
                    //numMultiplications += 1;
                }
                pp += _template->widthStep;
                pw += image->widthStep;
		}*/
	    /***/
	    float convolutionSum = CV_IMAGE_ELEM(convolution_response, float, y, x);
	    /***/

            windowNorm -= windowSum * windowSum / N;
            *q = (float)convolutionSum - patchMean * windowSum;
			/**n++ = *q;
			*d++ = sqrt(patchNorm * windowNorm);*/

            *q /= sqrt(patchNorm * windowNorm);
            q++;
        }
    }

	/*FILE *fo;
	fo = fopen("num.out","wb");
	fwrite(&convolution_response->width,sizeof(int),1,fo);
	fwrite(&convolution_response->height,sizeof(int),1,fo);
	fwrite(num,sizeof(float),convolution_response->height*convolution_response->width,fo);
	fclose(fo);
	fo = fopen("den.out","wb");
	fwrite(&convolution_response->width,sizeof(int),1,fo);
	fwrite(&convolution_response->height,sizeof(int),1,fo);
	fwrite(den,sizeof(float),convolution_response->height*convolution_response->width,fo);
	fclose(fo);

	free(num); free(den);*/

    /***/
    cvReleaseImage(&convolution_response);
    /***/

	//static int xena; char c[32]; sprintf(c,"resp_gold_%d.out",xena++);
	//FILE *fo = fopen(c,"wb");
	//fwrite(&response->width,sizeof(int),1,fo);
	//fwrite(&response->height,sizeof(int),1,fo);
	//for ( int r=0; r<response->height; ++r )
	//	fwrite(response->imageData + response->widthStep*r, sizeof(float), response->width, fo);
	//fclose(fo);

    //cerr << numMultiplications << " multiplications" << endl;
    svlCodeProfiler::toc(handle);
    return response;
}

IplImage *svlIntensityPatchDefinition::responseImage(vector<const IplImage *>& images, 
    vector<const IplImage *>& imageSums, vector<const IplImage *>& imageSumSqs) const
{
    assert((images.size() > (unsigned)_validChannel) &&
	(imageSums.size() > (unsigned)_validChannel) &&
	(imageSumSqs.size() > (unsigned)_validChannel));

    return responseImage(images[_validChannel], imageSums[_validChannel],
	imageSumSqs[_validChannel]);
}

vector<double> svlIntensityPatchDefinition::patchValues(const IplImage *image,
    const vector<CvPoint>& locations)
{		
    assert(image->depth == IPL_DEPTH_8U);


    // compute patch statistics
    float patchSum = 0.0;
    float patchNorm = 0.0;
    unsigned char *pp = (unsigned char *)_template->imageData;
    for (int y = 0; y < _template->height; y++) {
        for (int x = 0; x < _template->width; x++) {
            patchSum += (float)pp[x];
            patchNorm += (float)pp[x] * (float)pp[x];
        }
        pp += _template->widthStep;
    }
    float N = (float)(_template->width * _template->height);
    float patchMean = patchSum / N;
    patchNorm -= patchSum * patchMean;

    // precompute these in patch dictionary
    vector<double> values;
    values.reserve(locations.size());

    for (vector<CvPoint>::const_iterator it = locations.begin(); it != locations.end(); ++it) 
	{
        float maxVal = -numeric_limits<float>::max();
        for (int dy = _validRect.y; dy < _validRect.y + _validRect.height; dy++) {
            for (int dx = _validRect.x; dx < _validRect.x + _validRect.width; dx++) {
                const unsigned char *pw = &CV_IMAGE_ELEM(image, unsigned char, it->y + dy, it->x + dx);
                const unsigned char *pp = (const unsigned char *)_template->imageData;
                unsigned int windowSum = 0;
                unsigned int windowNorm = 0;
                unsigned int val = 0;
                for (int v = 0; v < _template->height; v++) {
                    for (int u = 0; u < _template->width; u++) {
                        windowSum += pw[u];
                        windowNorm += pw[u] * pw[u];
                        val += pp[u] * pw[u];
                    }
                    pp += _template->widthStep;
                    pw += image->widthStep;
                }

                float fWindowNorm = (float)windowNorm - (float)(windowSum * windowSum) / N;
                float fVal = (float)val - patchMean * (float)windowSum;
                fVal /= sqrt(patchNorm * fWindowNorm);
                if (fVal > maxVal && !isnan(fVal) && !isinf(fVal)) {
                    maxVal = fVal;
                }
            }
        }
        values.push_back((double)maxVal);
    }

    return values;
}

// svlDepthPatchDefinition class --------------------------------------------------

// initialize static data members
int svlDepthPatchDefinition::_maxDistance = 3;
int svlDepthPatchDefinition::_totalBins = 100;
CvSize svlDepthPatchDefinition::_windowSize;

svlDepthPatchDefinition::svlDepthPatchDefinition(svlPatchDictionary * parent) :
    svlPatchDefinition(parent)
{
    // do nothing
}

svlDepthPatchDefinition::svlDepthPatchDefinition(XMLNode& node, svlPatchDictionary * parent) :
  svlPatchDefinition(node, parent)
{
    // do nothing
	read(node);
}

svlDepthPatchDefinition::svlDepthPatchDefinition(const IplImage *t, const CvRect& rect, svlPatchDictionary * parent, int channel) :
  svlPatchDefinition(t, rect, parent, channel)
{
    // do nothing
}

svlDepthPatchDefinition::~svlDepthPatchDefinition()
{
    // do nothing
}

bool svlDepthPatchDefinition::read(XMLNode& node)
{	
	
    assert(!node.isEmpty());
	
    if (_template != NULL) {
        cvReleaseImage(&_template);
    }

    // read valid region
    if (node.getAttribute("validRect")) {
        vector<int> v;
        parseString<int>(node.getAttribute("validRect"), v);
        assert(v.size() == 4);
        _validRect = cvRect(v[0], v[1], v[2], v[3]);
    } else {
        _validRect = cvRect(0, 0, 0, 0);
    }

    // read channel
    if (node.getAttribute("channel")) {
        _validChannel = atoi(node.getAttribute("channel"));
    } else {
        _validChannel = 0;
    }

    // read patch size
    int w = 0;
    int h = 0;
    if (node.getAttribute("size")) {
        vector<int> v;
        parseString<int>(node.getAttribute("size"), v);
        assert(v.size() == 2);
        w = v[0]; h = v[1];
    }

    // read patch data
    if ((w > 0) && (h > 0)) {
        _template = cvCreateImage(cvSize(w, h), IPL_DEPTH_32F, 1);
        vector<float> v;
        parseString(node.getText(), v);
        assert(v.size() == (unsigned)(w * h));
        unsigned n = 0;
        for (unsigned y = 0; y < (unsigned)_template->height; y++) {
            float *p = &CV_IMAGE_ELEM(_template, float, y, 0);
            for (unsigned x = 0; x < (unsigned)_template->width; x++) {
                p[x] = v[n++];
            }
        }
    }
	
    return true;
}


vector<double> svlDepthPatchDefinition::patchValues(const IplImage *image, const vector<CvPoint>& windows)
{	
    SVL_ASSERT_MSG(image->nChannels == 1,
        "in svlDepthPatchDefinition::patchValues fn");

    vector<double> f(windows.size());
#if 0
    for (unsigned j = 0; j < windows.size(); j++) {
        //Compute response image for the patch
        // this is somethething that should be taken care of within svlPatchDictionary!

        //int handle = svlCodeProfiler::getHandle("svlDepthPatchDefinition::patchValues.iteration");        
        //svlCodeProfiler::tic(handle);   	

        IplImage *response = responseImage(_parent->_normalizedDepthPatches[j]);
        f[j] = patchValue(response, cvPoint(0,0));		        
        cvReleaseImage(&response);
        //svlCodeProfiler::toc(handle);
    }
#else
    CvRect roi = _validRect;
    roi.width += _template->width - 1;
    roi.height += _template->height - 1;
    if (!windows.empty()) {
      assert(_parent->_normalizedDepthPatches[0]);
        svlClipRect(roi, _parent->_normalizedDepthPatches[0]);
    }
    IplImage *subImage = cvCreateImage(cvSize(roi.width, roi.height), IPL_DEPTH_32F, 1);
    IplImage *response = cvCreateImage(cvSize(subImage->width - _template->width + 1,
        subImage->height - _template->height + 1), IPL_DEPTH_32F, 1);

    for (unsigned j = 0; j < windows.size(); j++) {
        // thread safe copy (instead of using cvSetImageROI)
        for (int y = 0; y < roi.height; y++) {
            float *p = (float *)&CV_IMAGE_ELEM(_parent->_normalizedDepthPatches[j],
                float, y + roi.y, roi.x);
            float *q = (float *)&CV_IMAGE_ELEM(subImage, float, y, 0);
            for (int x = 0; x < roi.width; x++) {
                q[x] = p[x];
            }
        }

        cvMatchTemplate(subImage, _template, response, CV_TM_CCORR_NORMED);

        double min_value, max_value;
        cvMinMaxLoc(response, &min_value, &max_value);
        f[j] = max_value;
    }
     
    cvReleaseImage(&response);
    cvReleaseImage(&subImage);
#endif
		
    return f;
}

//Takes the depth values of the template patch and computes a depth histogram
int** svlDepthPatchDefinition::computeDepthHistogram(const IplImage* patchTemplate , CvRect region)
{
	float averageDepth = 0.0;

	//Compute average depth of the template patch
	for( int i=region.y ; i< region.y + region.height ; i++ )
	{
		for( int j=region.x ; j< region.x + region.width ; j++)
		{			
			averageDepth += CV_IMAGE_ELEM(patchTemplate, float, i, j);
		}
	}	

		
	averageDepth /= region.height * region.width;

	//averageDepth = CV_IMAGE_ELEM(patchTemplate,float,region.y + region.height /2 , region.x + region.width /2);

	
	//printf("\n AD = %f", averageDepth );

	//Init the histogram
	int **histogram = new int*[32];

	for( int i=0 ; i<32 ; i++)
	{
		histogram[i] = new int[32];
	}

	//Compute average depth of the template patch
	for( int i=region.y ; i< region.y + region.height ; i++ )
	{
		for( int j=region.x ; j< region.x + region.width ; j++)
		{
			float patchValue = CV_IMAGE_ELEM(patchTemplate, float, i, j);
			//float normalizedDepth =  (patchValue - averageDepth) + svlDepthPatchDefinition::maxDistance;
			float normalizedDepth =  (patchValue - averageDepth) * 1000;

			histogram[i-region.y][j-region.x] = (int)normalizedDepth;
			
			//int index = ( (normalizedDepth * svlDepthPatchDefinition::totalBins) / ( 2.0 * svlDepthPatchDefinition::maxDistance )) ;

			//printf("\n %f %f %d",patchValue,normalizedDepth,index);
	
			//Check out of bound conditions
			//if( index < 0 )
			//	index = 0;
			//else if( index >= svlDepthPatchDefinition::totalBins )
			//	index = svlDepthPatchDefinition::totalBins -1;
		
			//histogram[index]++;

			//char gg;
			//scanf("%c",&gg);
		
		}
	}

	
	//printf("\nHistorgam ");
	//for( int i=0 ; i<svlDepthPatchDefinition::totalBins ; i++)
	//{
	//	printf(" %d " , histogram[i]);
	//}

	
	//char gg;
	//scanf("%c",&gg);
	
	//for( int i=0 ; i<3 ; i++)
	//{
	//	smoothHistogram(histogram,svlDepthPatchDefinition::totalBins);
		//}
	

	return histogram;

}

//Smooth histogram with a kernel of [1/3 1/3 1/3]
void svlDepthPatchDefinition::smoothHistogram(int *histogram, int totalLength)
{
   
	//Apply smoothing kernel
   for (int i = 1; i < totalLength - 1 ; i++)
   {
		histogram[i] = (histogram[i-1] + histogram[i] + histogram[i+1] ) / 3;  
   }

   //Handle first and last entry seperately
   histogram[0] = ( histogram[0] + histogram[1] ) / 2;
   histogram[totalLength-1] = ( histogram[totalLength-2] + histogram[totalLength-1] ) / 2;
}





//Computes the earth movers distances between the two given integer arrays (histograms)
int svlDepthPatchDefinition::computeEMD(int *a , int *b , int totalLength)
{
	float *af = new float[totalLength];
	float *bf = new float[totalLength];

	//Convert histograms to float types to use opencv emd function
	for( int i=0 ; i<totalLength ; i++)
	{
		af[i] = (float)a[i];
		bf[i] = (float)b[i];
	}

	float dist = cvCalcEMD(af,totalLength,bf,totalLength,0,CV_DIST_L1);

	delete []af;
	delete []bf;

	return (int)dist;
}


//Computes the euclidean distances between the two given integer arrays (histograms)
int svlDepthPatchDefinition::computeEuclideanDistance(int *a , int *b , int totalLength)
{

	float totalDistance = 0;

	for( int i=0 ; i<totalLength ; i++)
	{
            totalDistance += (a[i] - b[i]) * (a[i] - b[i]);
	}

	return (int)sqrt(totalDistance);
}

IplImage *svlDepthPatchDefinition::responseImage(const IplImage *image) const
{
	
//	int handle = svlCodeProfiler::getHandle("svlDepthPatchDefinition::responseImage");
 //   svlCodeProfiler::tic(handle);

    assert((_template != NULL) && (image != NULL));
    SVL_ASSERT_MSG(image->nChannels == 1,
		   "in svlDepthPatchDefinition::responseImage fn");
    
    IplImage *response = cvCreateImage(cvSize(image->width - _template->width + 1,
                                           image->height - _template->height + 1),
        IPL_DEPTH_32F, 1);
    cvMatchTemplate(image, _template, response, CV_TM_CCORR_NORMED);

    ////Compute histogram for the patch template
    //_histogram = computeDepthHistogram(_template , cvRect(0,0,_template->width,_template->height) );


    //for( int i= _template->height / 2; i< response->height - _template->height / 2 ; i++)
    //{
    //	for( int j= _template->width / 2 ; j< response->width - _template->width / 2 ; j++)
    //	{
    //		int *patchHistogram = computeDepthHistogram(image, cvRect(j -_template->width / 2
    //			,i - _template->height / 2,_template->width,_template->height));
    
    //		int distance = computeEuclideanDistance(_histogram,patchHistogram,svlDepthPatchDefinition::totalBins);
    //		
    //		//printf("\n%d %d %d %d",i,j,response->height,response->width);
    //		//cvSet2D(response,i,j,cvScalar(distance));
    //		
    //		CV_IMAGE_ELEM(response,float,i,j) = distance;	
    
    //		delete []patchHistogram;
    //
    //	}
    
    //}	
    
    //delete []_histogram;
    
    
    //	svlCodeProfiler::toc(handle);
  
    return response;
}

//Change to finding the minimum entry in response image because the lesser the distance the better
//double svlDepthPatchDefinition::patchValue(const IplImage *responseImage,  const CvPoint& location)
//{
//#if 0
//    assert((location.x + _validRect.x + _validRect.width <= responseImage->width) &&
//	   (location.y + _validRect.y + _validRect.height <= responseImage->height));
//#endif
//
//    float minScore = numeric_limits<float>::max();
//
//    float *p = &CV_IMAGE_ELEM(responseImage, float, location.y + _validRect.y,
//        location.x + _validRect.x);
//
//    size_t inc = responseImage->widthStep / sizeof(float);
//
//    for (int dy = 0; dy < _validRect.height; dy++) 
//	{
//		for (int dx = 0; dx < _validRect.width; dx++) 
//		{
//		    if (p[dx] < minScore) 
//			{
//				minScore = p[dx];
//			}
//		}
//
//		p += inc;
//    }
//
//    return (double)minScore;
//}



//*******************************************************************************************************************










// useful functions --------------------------------------------------


static inline IplImage *my_cvMake32F(const IplImage *image) {
  IplImage *result = cvCreateImage(cvSize(image->width, image->height),
				   IPL_DEPTH_32F, image->nChannels);
  cvConvertScale(image, result, 1.0/255.0);

  return result;      
}

static inline IplImage *my_cvMake8U(const IplImage *image, float scale = 1) {
  IplImage *result = cvCreateImage(cvSize(image->width, image->height),
				   IPL_DEPTH_8U, image->nChannels);
  cvConvertScale(image, result, 255.0*scale);

  return result;      
}

static void my_cvMult(IplImage *image, float factor) {
  SVL_ASSERT_MSG(image->depth == IPL_DEPTH_32F, "wrong depth in my_cvMult");

  cvScale(image, image, factor);

  return;


  float* data;
  int step;
  CvSize size;
  int x, y;
  int n = image->nChannels;
  
  cvGetRawData( image, (uchar**)&data, &step, &size);
  step /= sizeof(data[0]);
  
    for( y = 0; y < size.height; y++, data += step )
      for( x = 0; x < size.width * n; x++ )
	data[x] = data[x] * factor;
}

static void my_cvImageCrop(IplImage **im, CvRect region) {
  IplImage *image = *im;
  IplImage *result = cvCreateImage(cvSize(region.width, region.height),
				   image->depth, image->nChannels);
  cvSetImageROI(image, region);
  cvCopy(image, result);
  cvReleaseImage(im);
  *im = result;
}


// This is good, but kind of slow apparently for multi-channeled images especially.
// There is a different version below.

// the result will be an 8-bit unsigned image
static IplImage *my_cvEucDist(const IplImage *im) {
  //cerr << "entered my_cveucdist" << endl;

  assert(im->depth == IPL_DEPTH_32F);
  int n = im->nChannels;

  float* data;
  int step;
  CvSize size;
  int x, y, c;
  
  //cerr << "n is " << n << endl;

  cvGetRawData( im, (uchar**)&data, &step, &size);

  //cerr << "step: " << step << ", size: " << size.width << " " << size.height << " "
  //  << "sizeof(data[0]): " << sizeof(data[0]) << endl;

  step /= sizeof(data[0]);
  
  IplImage *result = cvCreateImage(cvSize(size.width, size.height),
				   IPL_DEPTH_32F, 1);
  
  //  cerr << "result's (width, height): (" << result->width << ", " << result->height << ")" << endl;

  float *newdata;
  int newstep;
  cvGetRawData(result, (uchar**)&newdata, &newstep);
  newstep /= sizeof(newdata[0]);
  
  if (n == 1) {
    for( y = 0; y < size.height; y++, data += step, newdata += newstep ) {
      for( x = 0; x < size.width; x++ ) {
	newdata[x] = fabs(data[x]);
	SVL_ASSERT_MSG((newdata[x] > -0.00001 && newdata[x] < 1.000001), "data is " << newdata[x]);
      }
    }
  }

  else {
    float tmp;
    
    for( y = 0; y < size.height; y++, data += step, newdata += newstep ) {
      //cerr << "y: " << y << endl;
      float *d = data;
      for( x = 0; x < size.width; x++) {
	//cerr << "x: " << x << endl;
	tmp = 0;
	for (c = 0; c < n; c++) {
	  tmp += (*d) * (*(d));
	  ++d;
	}
	//	tmp = sqrt(tmp) / norm;
	//assert(tmp > -0.000001);
	//assert(tmp < 1.0000001);
	//newdata[x] = tmp;
	newdata[x] = sqrt(tmp);
      }
    }
  }

  IplImage *result_8U = my_cvMake8U(result, 1.0f/sqrt((float)n));
  cvReleaseImage(&result);

  return result_8U;
}

static IplImage *my_cvEucDist(IplImage **im, int n, CvRect region) {
  assert(n > 0 && im != NULL && 
	 im[0]->depth == IPL_DEPTH_32F && im[0]->nChannels == 1);

  IplImage *result_8U;

  if (n == 1) {
    result_8U = my_cvEucDist(im[0]);
  } else {
    int width = im[0]->width;
    int height = im[0]->height;
    IplImage *result = cvCreateImage(cvSize(width, height),
				     IPL_DEPTH_32F, 1);
  
    cvCartToPolar(im[0], im[1], result);
    
    int index = 2;
    while (index < n) {
      cvCartToPolar(result, im[index++], result);
    }
    
    //  my_cvMult(result, cvInvSqrt(n));
    
#if 0
    int x, y;
    float *data;
    int step;
    cvGetRawData(result, (uchar**)&data, &step);
    step /= sizeof(data[0]);
    
    for( y = 0; y < height; y++, data += step ) {
      for( x = 0; x < width; x++) {
	if(! (data[x] > -0.000001 && data[x] < 1.00000001)) {
	  cerr << "(x, y): (" << x << ", " << y << ")" << endl;
	  cerr << "data[x]: " << data[x] << endl;
	  for (int i = 0; i < n; i++)
	    cerr << "in image " << i << ": " << cvGetReal2D(im[i], y, x) << endl;
	  assert(false);
	}
      }
    }
#endif

    result_8U = my_cvMake8U(result, cvInvSqrt((float)n));
    
    cvReleaseImage(&result);
  }

  my_cvImageCrop(&result_8U, region);

  return result_8U;
}









// svlAbsPatchDefinition class --------------------------------------------------

svlAbsPatchDefinition::svlAbsPatchDefinition(svlPatchDictionary * parent, svlAbsPatchType tp) :
  svlPatchDefinition(parent), _masks(NULL), _refs(NULL), _responses(NULL)
{
  svlAbsInitialize(tp);
}

svlAbsPatchDefinition::svlAbsPatchDefinition(XMLNode& node, svlPatchDictionary * parent) :
  svlPatchDefinition(parent), _masks(NULL), _refs(NULL), _responses(NULL)
{
  read(node);
}

// be able to also include the masks in here so don't have to be initialized every time
svlAbsPatchDefinition::svlAbsPatchDefinition(const IplImage *t, const CvRect& rect, svlPatchDictionary * parent, int channel,
					     svlAbsPatchType tp) :
  svlPatchDefinition(t, rect, parent, channel), _masks(NULL), _refs(NULL), _responses(NULL)
{
  //cerr << "entering svlAbsPatchDefinition()" << endl;

  if (_template->depth != IPL_DEPTH_32F) {
    IplImage *temp = my_cvMake32F(_template);
    cvReleaseImage(&_template);
    _template = temp;
  }

  svlAbsInitialize(tp);

  // compute the expected responses based on the patch information
  _responses = maskResponseImages(_template);
  //cerr << "exiting svlAbsPatchDefinition()" << endl;
}

svlAbsPatchDefinition::~svlAbsPatchDefinition()
{
  if (_responses) {
    for (int i = 0; i < _nmasks; i++) {
      if (_responses[i]) {
	cvReleaseImage(_responses + i);
	_responses[i] = NULL;
      }
    }
    delete[] _responses;
    _responses = NULL;
  }

  if (_masks) {
    for (int i = 0; i < _nmasks; i++) {
      if (_masks[i]) {
	//cerr << "Trying to release mask" << i << endl;
	cvReleaseMat(_masks + i);
	//cerr << "success" << endl;
	_masks[i] = NULL;
      }
    }
    delete[] _masks;
    //_masks = NULL;
  }

  if (_refs) {
    delete[] _refs;
    _refs = NULL;
  }
}


void svlAbsPatchDefinition::svlAbsInitialize(svlAbsPatchType t) {
  _type = t;
  _responses = NULL;

  _responseOffsetBR =_responseOffsetTL = cvPoint(0, 0);

  svlAbsPatchMaskCreator cr;

  switch(t) {
  case SVL_ABS_NE:
  case SVL_ABS_E:
  case SVL_ABS_SE:
  case SVL_ABS_S:
    _nmasks = cr.numMasksRequiredDirectional(); break;
  case SVL_ABS_INIT:
    _nmasks = cr.numMasksRequired(1) + cr.numMasksRequired(2) + cr.numMasksRequired(3); break;
  case SVL_ABS_ONE:
    _nmasks = cr.numMasksRequired(1); break;
  case SVL_ABS_TWO:
    _nmasks = cr.numMasksRequired(2); break;
  case SVL_ABS_FOUR:
    _nmasks = cr.numMasksRequired(4); break;
  default:
    SVL_ASSERT_MSG(false, "this type of AbsPatch is undefined and thus can't be initialized");
  }

  _masks = new CvMat*[_nmasks];
  _refs = new CvPoint[_nmasks];

  int num = 0;
  switch(t) {
  case SVL_ABS_NE:
  case SVL_ABS_E:
  case SVL_ABS_SE:
  case SVL_ABS_S:
    cr.createDirectional((int)t, _masks, _refs, _responseOffsetTL, _responseOffsetBR); break;
  case SVL_ABS_INIT:
    num += cr.create(1, _masks+num, _refs+num, _responseOffsetTL, _responseOffsetBR);
    num += cr.create(2, _masks+num, _refs+num, _responseOffsetTL, _responseOffsetBR);
    num += cr.create(3, _masks+num, _refs+num, _responseOffsetTL, _responseOffsetBR);
  case SVL_ABS_ONE:
    cr.create(1, _masks, _refs, _responseOffsetTL, _responseOffsetBR); break;
  case SVL_ABS_TWO:
    cr.create(2, _masks, _refs, _responseOffsetTL, _responseOffsetBR); break;
  case SVL_ABS_FOUR:
    cr.create(4, _masks, _refs, _responseOffsetTL, _responseOffsetBR); break;
  default:
    SVL_ASSERT_MSG(false, "this type of AbsPatch is undefined and thus can't be initialized");
  }

}


vector<double> svlAbsPatchDefinition::patchValues(const IplImage *image, 
    const vector<CvPoint>& locations) {
  IplImage *response = responseImage(image);

  vector<double> result(locations.size());
  for(unsigned i = 0; i < locations.size(); i++) {
    result[i] = patchValue(response, locations[i]);
  }
  cvReleaseImage(&response);
  return result;
}

vector<double> svlAbsPatchDefinition::patchValues(IplImage **imResponses, 
    const vector<CvPoint>& locations) {
  IplImage *response = responseImage(imResponses);

  vector<double> result(locations.size());
  for(unsigned i = 0; i < locations.size(); i++) {
    result[i] = patchValue(response, locations[i]);
  }
  cvReleaseImage(&response);
  return result;
}


IplImage *svlAbsPatchDefinition::responseImage(IplImage **imResponses) const {
  SVL_ASSERT_MSG(imResponses[0]->depth == _responses[0]->depth, "wrong depth in responseImage(imResponses)");

  int width = imResponses[0]->width;
  int height = imResponses[0]->height;

  int patch_width = _responses[0]->width;
  int patch_height = _responses[0]->height;

  // this is actually the same as before, image->width - _template->width + 1
  int templ_match_width = width - patch_width + 1;
  int templ_match_height = height - patch_height + 1;

  //cerr << "imResponses are " << width << " x " << height
  //   << ", and _responses are " << patch_width << " x " << patch_height << endl;
  
  IplImage *result = cvCreateImage(cvSize(templ_match_width, templ_match_height), IPL_DEPTH_32F, 1);
  cvSetZero(result);

  IplImage *tmp = cvCreateImage(cvSize(templ_match_width, templ_match_height),
				IPL_DEPTH_32F, 1);

  for (int i = 0; i < _nmasks; i++) {
    cvMatchTemplate(imResponses[i], _responses[i], tmp, CV_TM_CCOEFF_NORMED);
    cvAdd(tmp, result, result);
  }

  my_cvMult(result, 1.0f/(float)_nmasks);
  cvReleaseImage(&tmp);

  return result;
}

IplImage *svlAbsPatchDefinition::responseImage(const IplImage *image) const {

  IplImage **imResponses;

  if (image->depth != IPL_DEPTH_32F) {
    IplImage *im = my_cvMake32F(image);
    imResponses = maskResponseImages(im);
    cvReleaseImage(&im);
  } else {
    imResponses = maskResponseImages(image);
  }

  IplImage *result = responseImage(imResponses);

  deallocateResponseImages(imResponses);

  return result;
}

void svlAbsPatchDefinition::deallocateResponseImages(IplImage **&imResponses) const {
  for (int i = 0; i < _nmasks; i++)
    cvReleaseImage(imResponses+i);
  delete[] imResponses;

  imResponses = NULL;
}

// image should be 32_F image
IplImage **svlAbsPatchDefinition::maskResponseImages(const IplImage *im) const {
  IplImage *image;

  if(im->depth != IPL_DEPTH_32F) {
    image = my_cvMake32F(im);
  } else {
    image = cvCloneImage(im);
  }

  CvRect region;
  region.x = _responseOffsetTL.x;
  region.y = _responseOffsetTL.y;
  region.width = image->width - _responseOffsetTL.x - _responseOffsetBR.x;
  region.height = image->height - _responseOffsetTL.y - _responseOffsetBR.y;

  /*  cerr << "In maskResponseImages" << endl;
  cerr << "\t_responseOffsetTL.x " << _responseOffsetTL.x << endl;
  cerr << "\t_responseOffsetTL.y " << _responseOffsetTL.y << endl;
  cerr << "\t_responseOffsetBR.x " << _responseOffsetBR.x << endl;
  cerr << "\t_responseOffsetBR.y " << _responseOffsetBR.y << endl;
  cerr << "\timage->width " << image->width << endl;
  cerr << "\timage->height " << image->height << endl;
  cerr << "\tim->width " << im->width << endl;
  cerr << "\tim->height " << im->height << endl;
  */

  IplImage **responses = new IplImage*[_nmasks];

  if (image->nChannels == 1) {
    IplImage *resp = cvCreateImage(cvSize(image->width, image->height),
				   image->depth, 1);

    for (int i = 0; i < _nmasks; i++) {
      svlConvolution conv(_masks[i], _refs[i]);

      cvResetImageROI(resp);
      conv.filter(image, resp, true);
      cvSetImageROI(resp, region);
      responses[i] = my_cvEucDist(resp);
    }

    cvReleaseImage(&resp);
  } 

  else if (image->nChannels == 2) {
    IplImage *imageChannels[2];
    IplImage *resp[2];
    for (int c = 0; c < 2; c++) {
      imageChannels[c] = cvCreateImage(cvSize(image->width, image->height), 
				       image->depth, 1);
      resp[c] = cvCreateImage(cvSize(image->width, image->height),
			      image->depth, 1);
    }
    cvSplit(image, imageChannels[0], imageChannels[1], NULL, NULL);
    
    for (int i = 0; i < _nmasks; i++) {
      svlConvolution conv(_masks[i], _refs[i]);

      for (int c = 0; c < 2; c++) {
	//cvResetImageROI(resp[c]);
	conv.filter(imageChannels[c], resp[c], true);
	//cvSetImageROI(resp[c], region);
      }

      //cvResetImageROI(fullResponse);
      //cvMerge(resp[0], resp[1], resp[2], NULL, fullResponse);
      //cvSetImageROI(fullResponse, region);
      //responses[i] = my_cvEucDist(fullResponse);

      responses[i] = my_cvEucDist(resp, 2, region);

      //cerr << "euc dist computed" << endl;
    } 

    for (int c = 0; c < 2; c++) {
      cvReleaseImage(&(imageChannels[c]));
      cvReleaseImage(&(resp[c]));
    }

  }

  else if (image->nChannels == 3) {
    //IplImage *fullResponse = cvCreateImage(cvSize(image->width, image->height),
    //				   image->depth, 3);

    IplImage *imageChannels[3];
    IplImage *resp[3];
    for (int c = 0; c < 3; c++) {
      imageChannels[c] = cvCreateImage(cvSize(image->width, image->height), 
				       image->depth, 1);
      resp[c] = cvCreateImage(cvSize(image->width, image->height),
			      image->depth, 1);
    }
    cvSplit(image, imageChannels[0], imageChannels[1], imageChannels[2], NULL);
    
    for (int i = 0; i < _nmasks; i++) {
      svlConvolution conv(_masks[i], _refs[i]);

      for (int c = 0; c < 3; c++) {
	//cvResetImageROI(resp[c]);
	conv.filter(imageChannels[c], resp[c], true);
	//cvSetImageROI(resp[c], region);
      }

      //cvResetImageROI(fullResponse);
      //cvMerge(resp[0], resp[1], resp[2], NULL, fullResponse);
      //cvSetImageROI(fullResponse, region);
      //responses[i] = my_cvEucDist(fullResponse);

      responses[i] = my_cvEucDist(resp, 3, region);

      //cerr << "euc dist computed" << endl;
    } 

    for (int c = 0; c < 3; c++) {
      cvReleaseImage(&(imageChannels[c]));
      cvReleaseImage(&(resp[c]));
    }
    //cvReleaseImage(&fullResponse);
  }

  else {
    assert(false);
  }

  cvReleaseImage(&image);

  return responses;
}


// The read and write functions really shouldn't be dealing with the template,
// which, really, shouldn't be stored in the first place. 

bool svlAbsPatchDefinition::read(XMLNode &node)
{		
  assert(!node.isEmpty());
  if (_template != NULL) {
    cvReleaseImage(&_template);
  }
  if (_responses) {
    for (int i = 0; i < _nmasks; i++) {
      if (_responses[i])
	cvReleaseImage(_responses + i);
    }
    delete[] _responses;
  }
  

    // read valid region
    if (node.getAttribute("validRect")) {
        vector<int> v;
        parseString<int>(node.getAttribute("validRect"), v);
        assert(v.size() == 4);
        _validRect = cvRect(v[0], v[1], v[2], v[3]);
    } else {
        _validRect = cvRect(0, 0, 0, 0);
    }

    // read channel
    if (node.getAttribute("channel")) {
        _validChannel = atoi(node.getAttribute("channel"));
    } else {
        _validChannel = 0;
    }

    // read subtype
    if (node.getAttribute("subtype")) {
      _type = (svlAbsPatchType)atoi(node.getAttribute("subtype"));
    } else {
      _type = (svlAbsPatchType)0;
    }

    // read patch size
    int w = 0;
    int h = 0;
    int nc = 0;
    if (node.getAttribute("size")) {
        vector<int> v;
        parseString<int>(node.getAttribute("size"), v);
        assert(v.size() == 3);
        w = v[0]; h = v[1]; nc = v[2];
    }

    // read patch data
    if ((w > 0) && (h > 0) && (nc > 0)) {
        _template = cvCreateImage(cvSize(w, h), IPL_DEPTH_32F, nc);
        vector<float> v;
        parseString<float>(node.getText(), v);
        assert(v.size() == (unsigned)(w * h * nc));
        unsigned n = 0;
        for (unsigned y = 0; y < (unsigned)_template->height; y++) {
	  float *p = (float *)&_template->imageData[y * _template->widthStep];
            for (unsigned x = 0; x < (unsigned)_template->width*_template->nChannels; x++) {
                p[x] = v[n++];
            }
        }
    }

    // read color space and store it with the image for
    // future reference (I don't think we actually need to know this
    // except for display purposes)
    //if (node.getAttribute("color")) {
    //sprintf(_template.colorModel, node.getAttribute("color"));
    //} 

    svlAbsInitialize(_type);
    _responses = maskResponseImages(_template);
	
    return true;
}

bool svlAbsPatchDefinition::write(ostream& os)
{
    assert(_template != NULL);
    assert(_template->depth == IPL_DEPTH_32F);

    os << "  <PatchDefinition type=\"abs\"\n"
       << "    validRect=\"" << _validRect.x << " " << _validRect.y << " "
       << _validRect.width << " " << _validRect.height << "\"\n"
       << "    channel=\"" << _validChannel << "\"\n"
       << "    subtype=\"" << _type << "\"\n"
       << "    size=\"" << _template->width << " "
       << _template->height << " " << _template->nChannels << "\">\n";
    
    for (unsigned y = 0; y < (unsigned)_template->height; y++) 
      {
        os << "   ";
        for (unsigned x = 0; x < (unsigned)_template->width * _template->nChannels; x++) 
	  {
	    os << " " << CV_IMAGE_ELEM(_template, float, y, x);
	  }
        os << "\n";
      }
    
    os << "  </PatchDefinition>\n";

    return true;
}



bool isImageUniform(IplImage *img, double min_delta ) {
  SVL_ASSERT_MSG(img, "Checking for whether a NULL patch is uniform");

  bool uniform = false;
  for (int i = 1; i <= img->nChannels; i++) {
    cvSetImageCOI(img, i);
    double minVal, maxVal;
    cvMinMaxLoc(img, &minVal, &maxVal);

    // if at least one channel is uniform, the entire thing is considered uniform
    uniform = (uniform) || (maxVal - minVal <= min_delta);

    if (uniform)
      break;
  }
  cvSetImageCOI(img, 0);
  return uniform;
}

bool isImageAlmostUniform(IplImage *img)
{
  // image can have ROI set

  SVL_ASSERT_MSG(img, "Checking for whether a NULL patch is uniform");

  double min_delta = 0;
  if (img->depth == IPL_DEPTH_8U) {
    min_delta = 32.0; 
  } else if (img->depth == IPL_DEPTH_32F) {
    min_delta = 0.125;
  } else {
    SVL_ASSERT_MSG(false, "An image of depth " << img->depth
		   << " is not currently supported by isPatchUniform");
  }

  return isImageUniform(img, min_delta);
}
